from abc import abstractmethod
from .. import _ffi as ffi, ImportType, ExportType, WasmtimeError, Storelike
from ._resources import ResourceAny, ResourceHost
from ._resource_type import ResourceType
from dataclasses import dataclass
import ctypes
from ctypes import byref, POINTER
from typing import Any, Optional, Dict, List, Union, Tuple, Set
from wasmtime import Managed, Engine


class ComponentType(Managed["ctypes._Pointer[ffi.wasmtime_component_type_t]"]):
    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `ComponentType`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_type_t]") -> None:
        ffi.wasmtime_component_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_type_t]") -> "ComponentType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_type_t)):
            raise TypeError("wrong pointer type")
        ty: "ComponentType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    def imports(self, engine: Engine) -> Dict[str, "ComponentItem"]:
        """
        Returns a dictionary of the imports of this component type.
        """
        n = ffi.wasmtime_component_type_import_count(self.ptr(), engine.ptr())
        items = {}
        for i in range(n):
            name_ptr = ctypes.POINTER(ctypes.c_char)()
            name_len = ctypes.c_size_t()
            item = ffi.wasmtime_component_item_t()
            found = ffi.wasmtime_component_type_import_nth(self.ptr(),
                                                           engine.ptr(),
                                                           i,
                                                           byref(name_ptr),
                                                           byref(name_len),
                                                           byref(item))
            assert(found)
            name = ctypes.string_at(name_ptr, name_len.value).decode('utf-8')
            items[name] = component_item_from_ptr(item)
        return items

    def exports(self, engine: Engine) -> Dict[str, "ComponentItem"]:
        """
        Returns a dictionary of the exports of this component type.
        """
        n = ffi.wasmtime_component_type_export_count(self.ptr(), engine.ptr())
        items = {}
        for i in range(n):
            name_ptr = ctypes.POINTER(ctypes.c_char)()
            name_len = ctypes.c_size_t()
            item = ffi.wasmtime_component_item_t()
            found = ffi.wasmtime_component_type_export_nth(self.ptr(),
                                                           engine.ptr(),
                                                           i,
                                                           byref(name_ptr),
                                                           byref(name_len),
                                                           byref(item))
            assert(found)
            name = ctypes.string_at(name_ptr, name_len.value).decode('utf-8')
            items[name] = component_item_from_ptr(item)
        return items


class ComponentInstanceType(Managed["ctypes._Pointer[ffi.wasmtime_component_instance_type_t]"]):
    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `ComponentInstanceType`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_instance_type_t]") -> None:
        ffi.wasmtime_component_instance_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_instance_type_t]") -> "ComponentInstanceType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_instance_type_t)):
            raise TypeError("wrong pointer type")
        ty: "ComponentInstanceType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    def exports(self, engine: Engine) -> Dict[str, "ComponentItem"]:
        """
        Returns a dictionary of the exports of this component instance type.
        """
        n = ffi.wasmtime_component_instance_type_export_count(self.ptr(), engine.ptr())
        items = {}
        for i in range(n):
            name_ptr = ctypes.POINTER(ctypes.c_char)()
            name_len = ctypes.c_size_t()
            item = ffi.wasmtime_component_item_t()
            found = ffi.wasmtime_component_instance_type_export_nth(self.ptr(),
                                                           engine.ptr(),
                                                           i,
                                                           byref(name_ptr),
                                                           byref(name_len),
                                                           byref(item))
            name = ctypes.string_at(name_ptr, name_len.value).decode('utf-8')
            items[name] = component_item_from_ptr(item)
            assert(found)
        return items


class ModuleType(Managed["ctypes._Pointer[ffi.wasmtime_module_type_t]"]):
    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `ModuleType`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_module_type_t]") -> None:
        ffi.wasmtime_module_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_module_type_t]") -> "ModuleType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_module_type_t)):
            raise TypeError("wrong pointer type")
        ty: "ModuleType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    def imports(self, engine: Engine) -> List[ImportType]:
        """
        Returns a list of the imports of this module type.
        """
        n = ffi.wasmtime_module_type_import_count(self.ptr(), engine.ptr())
        items = []
        for i in range(n):
            item = ffi.wasmtime_module_type_import_nth(self.ptr(),
                                                       engine.ptr(),
                                                       i)
            items.append(ImportType._from_ptr(item))
        return items

    def exports(self, engine: Engine) -> List[ExportType]:
        """
        Returns a list of the exports of this module type.
        """
        n = ffi.wasmtime_module_type_export_count(self.ptr(), engine.ptr())
        items = []
        for i in range(n):
            item = ffi.wasmtime_module_type_export_nth(self.ptr(),
                                                       engine.ptr(),
                                                       i)
            items.append(ExportType._from_ptr(item))
        return items


class FuncType(Managed["ctypes._Pointer[ffi.wasmtime_component_func_type_t]"]):
    _owner: Any

    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `FuncType`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_func_type_t]") -> None:
        if self._owner is None:
            ffi.wasmtime_component_func_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_func_type_t]", owner: Any = None) -> "FuncType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_func_type_t)):
            raise TypeError("wrong pointer type")
        ty: "FuncType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        ty._owner = owner
        return ty

    @property
    def params(self) -> List[Tuple[str, 'ValType']]:
        """
        Returns the parameter types of this component function type.
        """
        n = ffi.wasmtime_component_func_type_param_count(self.ptr())
        items = []
        for i in range(n):
            name_ptr = ctypes.POINTER(ctypes.c_char)()
            name_len = ctypes.c_size_t()
            valtype_ptr = ffi.wasmtime_component_valtype_t()
            found = ffi.wasmtime_component_func_type_param_nth(self.ptr(),
                                                               i,
                                                               byref(name_ptr),
                                                               byref(name_len),
                                                               byref(valtype_ptr))
            assert(found)
            name = ctypes.string_at(name_ptr, name_len.value).decode('utf-8')
            valtype = valtype_from_ptr(valtype_ptr)
            items.append((name, valtype))
        return items

    @property
    def result(self) -> Optional['ValType']:
        """
        Returns the result type of this component function type, if any.
        """
        valtype_ptr = ffi.wasmtime_component_valtype_t()
        has_result = ffi.wasmtime_component_func_type_result(self.ptr(), byref(valtype_ptr))
        if not has_result:
            return None
        return valtype_from_ptr(valtype_ptr)


class ValType:
    @abstractmethod
    def add_classes(self, s: Set[type]) -> None:
        """
        Returns the python class that is created by `convert_from_c` and
        accepted by `convert_to_c`
        """
        pass

    @abstractmethod
    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        """
        Converts `val` to this type and stores it in `ptr`
        """
        pass

    @abstractmethod
    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        """
        Converts `val` to Python
        """
        pass


@dataclass
class Bool(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(bool)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, bool):
            raise TypeError("expected bool for Bool type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_BOOL
        ptr.contents.of.boolean = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_BOOL.value)
        return bool(c.of.boolean)


@dataclass
class S8(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(int)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, int):
            raise TypeError("expected int for S8 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_S8
        ptr.contents.of.s8 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_S8.value)
        return int(c.of.s8)


@dataclass
class S16(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(int)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, int):
            raise TypeError("expected int for S16 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_S16
        ptr.contents.of.s16 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_S16.value)
        return int(c.of.s16)


@dataclass
class S32(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(int)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, int):
            raise TypeError("expected int for S32 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_S32
        ptr.contents.of.s32 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_S32.value)
        return int(c.of.s32)


@dataclass
class S64(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(int)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, int):
            raise TypeError("expected int for S64 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_S64
        ptr.contents.of.s64 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_S64.value)
        return int(c.of.s64)


@dataclass
class U8(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(int)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, int):
            raise TypeError("expected int for U8 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_U8
        ptr.contents.of.u8 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_U8.value)
        return int(c.of.u8)


@dataclass
class U16(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(int)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, int):
            raise TypeError("expected int for U16 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_U16
        ptr.contents.of.u16 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_U16.value)
        return int(c.of.u16)


@dataclass
class U32(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(int)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, int):
            raise TypeError("expected int for U32 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_U32
        ptr.contents.of.u32 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_U32.value)
        return int(c.of.u32)


@dataclass
class U64(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(int)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, int):
            raise TypeError("expected int for U64 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_U64
        ptr.contents.of.u64 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_U64.value)
        return int(c.of.u64)


@dataclass
class F32(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(float)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, float):
            raise TypeError("expected float for F32 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_F32
        ptr.contents.of.f32 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_F32.value)
        return float(c.of.f32)


@dataclass
class F64(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(float)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, float):
            raise TypeError("expected float for F64 type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_F64
        ptr.contents.of.f64 = val

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_F64.value)
        return float(c.of.f64)


@dataclass
class Char(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(str)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, str) or len(val) != 1:
            raise TypeError("expected single-character string for Char type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_CHAR
        ptr.contents.of.character = ord(val)

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_CHAR.value)
        return chr(c.of.character)



@dataclass
class String(ValType):
    def add_classes(self, s: Set[type]) -> None:
        s.add(str)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, str):
            raise TypeError("expected string type")
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_STRING
        ptr.contents.of.string = ffi.str_to_capi(val)

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_STRING.value)
        ret = ffi.to_str(c.of.string)
        ffi.wasm_byte_vec_delete(byref(c.of.string))
        return ret


@dataclass
class ErrorContext(ValType):
    def add_classes(self, s: Set[type]) -> None:
        raise NotImplementedError("ErrorContext conversion not implemented yet")

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        raise NotImplementedError("ErrorContext conversion not implemented yet")

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        raise NotImplementedError("ErrorContext conversion not implemented yet")


class ListType(Managed["ctypes._Pointer[ffi.wasmtime_component_list_type_t]"], ValType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_list_type_t]") -> None:
        ffi.wasmtime_component_list_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_list_type_t]") -> "ListType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_list_type_t)):
            raise TypeError("wrong pointer type")
        ty: "ListType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def element(self) -> 'ValType':
        """
        Returns the element type of this list type.
        """
        valtype_ptr = ffi.wasmtime_component_valtype_t()
        ffi.wasmtime_component_list_type_element(self.ptr(), byref(valtype_ptr))
        return valtype_from_ptr(valtype_ptr)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ListType):
            return False
        return ffi.wasmtime_component_list_type_equal(self.ptr(), other.ptr())

    def _is_bytes(self) -> bool:
        return isinstance(self.element, U8)

    def add_classes(self, s: Set[type]) -> None:
        if self._is_bytes():
            s.add(bytes)
        else:
            s.add(list)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        element = self.element
        if self._is_bytes():
            if not isinstance(val, bytes):
                raise TypeError("expected bytes value")
        else:
            if not isinstance(val, list):
                raise TypeError("expected list value")
        assert(isinstance(val, (bytes, list))) # here for mypy
        raw = ffi.wasmtime_component_vallist_t()
        ffi.wasmtime_component_vallist_new_uninit(raw, len(val))
        i = 0
        cleanup = True
        try:
            for e in val:
                element.convert_to_c(store, e, ctypes.pointer(raw.data[i]))
                i += 1
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_LIST
            ptr.contents.of.list = raw
            cleanup = False
        finally:
            if cleanup:
                for j in range(i, len(val)):
                    raw.data[j].kind = ffi.WASMTIME_COMPONENT_BOOL
                    raw.data[j].of.boolean = False
                ffi.wasmtime_component_vallist_delete(byref(raw))


    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_LIST.value)
        try:
            ret = []
            element = self.element
            for i in range(c.of.record.size):
                raw_field = c.of.tuple.data[i]
                ret.append(element.convert_from_c(raw_field))
                raw_field.kind = ffi.WASMTIME_COMPONENT_BOOL
                raw_field.of.boolean = False
            if self._is_bytes():
                return bytes(ret)
            return ret
        finally:
            ffi.wasmtime_component_valtuple_delete(byref(c.of.tuple))


class RecordType(Managed["ctypes._Pointer[ffi.wasmtime_component_record_type_t]"], ValType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_record_type_t]") -> None:
        ffi.wasmtime_component_record_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_record_type_t]") -> "RecordType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_record_type_t)):
            raise TypeError("wrong pointer type")
        ty: "RecordType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def fields(self) -> List[Tuple[str, 'ValType']]:
        """
        Returns the fields of this record type.
        """
        n = ffi.wasmtime_component_record_type_field_count(self.ptr())
        items = []
        for i in range(n):
            name_ptr = ctypes.POINTER(ctypes.c_char)()
            name_len = ctypes.c_size_t()
            valtype_ptr = ffi.wasmtime_component_valtype_t()
            found = ffi.wasmtime_component_record_type_field_nth(self.ptr(),
                                                                 i,
                                                                 byref(name_ptr),
                                                                 byref(name_len),
                                                                 byref(valtype_ptr))
            assert(found)
            name = ctypes.string_at(name_ptr, name_len.value).decode('utf-8')
            valtype = valtype_from_ptr(valtype_ptr)
            items.append((name, valtype))
        return items

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RecordType):
            return False
        return ffi.wasmtime_component_record_type_equal(self.ptr(), other.ptr())

    def add_classes(self, s: Set[type]) -> None:
        s.add(Record)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        fields = self.fields
        raw = ffi.wasmtime_component_valrecord_t()
        ffi.wasmtime_component_valrecord_new_uninit(raw, len(fields))
        i = 0
        cleanup = True
        try:
            for name, ty in fields:
                ty.convert_to_c(store, getattr(val, name), ctypes.pointer(raw.data[i].val))
                raw.data[i].name = ffi.str_to_capi(name)
                i += 1
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_RECORD
            ptr.contents.of.record = raw
            cleanup = False
        finally:
            if cleanup:
                for j in range(i, len(fields)):
                    raw.data[j].val.kind = ffi.WASMTIME_COMPONENT_BOOL
                    raw.data[j].val.of.boolean = False
                ffi.wasmtime_component_valrecord_delete(byref(raw))


    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_RECORD.value)
        try:
            ret = Record()
            fields = self.fields
            for i, (name, ty) in zip(range(c.of.record.size), fields):
                raw_field = c.of.record.data[i]
                raw_name = ffi.to_str(raw_field.name)
                assert(raw_name == name)
                val = ty.convert_from_c(raw_field.val)
                setattr(ret, name, val)
                raw_field.val.kind = ffi.WASMTIME_COMPONENT_BOOL
                raw_field.val.of.boolean = False
            return ret
        finally:
            ffi.wasmtime_component_valrecord_delete(c.of.record)


class Record:
    def __eq__(self, other: Any) -> bool:
        for key, value in self.__dict__.items():
            if not hasattr(other, key) or getattr(other, key) != value:
                return False
        return True


class TupleType(Managed["ctypes._Pointer[ffi.wasmtime_component_tuple_type_t]"], ValType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_tuple_type_t]") -> None:
        ffi.wasmtime_component_tuple_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_tuple_type_t]") -> "TupleType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_tuple_type_t)):
            raise TypeError("wrong pointer type")
        ty: "TupleType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def elements(self) -> List['ValType']:
        """
        Returns the element types of this tuple type.
        """
        n = ffi.wasmtime_component_tuple_type_types_count(self.ptr())
        items = []
        for i in range(n):
            valtype_ptr = ffi.wasmtime_component_valtype_t()
            found = ffi.wasmtime_component_tuple_type_types_nth(self.ptr(),
                                                                i,
                                                                byref(valtype_ptr))
            assert(found)
            valtype = valtype_from_ptr(valtype_ptr)
            items.append(valtype)
        return items

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TupleType):
            return False
        return ffi.wasmtime_component_tuple_type_equal(self.ptr(), other.ptr())

    def add_classes(self, s: Set[type]) -> None:
        s.add(tuple)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        elements = self.elements
        if not isinstance(val, tuple):
            raise TypeError("expected tuple type")
        if len(val) != len(elements):
            raise TypeError("tuple length mismatch")
        raw = ffi.wasmtime_component_valtuple_t()
        ffi.wasmtime_component_valtuple_new_uninit(raw, len(elements))
        i = 0
        cleanup = True
        try:
            for ty, element in zip(elements, val):
                ty.convert_to_c(store, element, ctypes.pointer(raw.data[i]))
                i += 1
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_TUPLE
            ptr.contents.of.tuple = raw
            cleanup = False
        finally:
            if cleanup:
                for j in range(i, len(elements)):
                    raw.data[j].kind = ffi.WASMTIME_COMPONENT_BOOL
                    raw.data[j].of.boolean = False
                ffi.wasmtime_component_valtuple_delete(byref(raw))


    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_TUPLE.value)
        try:
            ret: List[Any] = []
            for i, ty in zip(range(c.of.record.size), self.elements):
                raw_field = c.of.tuple.data[i]
                ret.append(ty.convert_from_c(raw_field))
                raw_field.kind = ffi.WASMTIME_COMPONENT_BOOL
                raw_field.of.boolean = False
            return tuple(ret)
        finally:
            ffi.wasmtime_component_valtuple_delete(byref(c.of.tuple))


@dataclass
class Variant:
    tag: str
    payload: Optional[Any] = None


case_test_cache: Set[type] = set()


class VariantLikeType:
    @abstractmethod
    def _cases(self) -> List[Tuple[str, Optional['ValType']]]:
        pass

    def add_classes(self, s: Set[type]) -> None:
        if self._tagged():
            s.add(Variant)
        else:
            for _, ty in self._cases():
                if ty is None:
                    s.add(object)
                else:
                    ty.add_classes(s)

    def _tagged(self) -> bool:
        t: Set[type] = set()
        for _, ty in self._cases():
            case: Set[type] = set()
            if ty is None:
                case.add(object)
            else:
                ty.add_classes(case)
            if len(case & t) != 0:
                return True
            t |= case
        return False

    def _lower(self, store: Storelike, val: Any) -> Tuple[str, Optional['ctypes._Pointer[ffi.wasmtime_component_val_t]']]:
        tagged = self._tagged()
        if tagged and not isinstance(val, Variant):
            raise TypeError("expected Variant type")
        for name, ty in self._cases():
            if tagged:
                if name != val.tag:
                    continue
            elif ty is None:
                if val is not None:
                    continue
            else:
                case_test_cache.clear()
                ty.add_classes(case_test_cache)
                if not isinstance(val, tuple(case_test_cache)):
                    continue

            if ty is None:
                return (name, None)
            raw = ffi.wasmtime_component_val_t()
            if tagged:
                payload = val.payload
            else:
                payload = val
            ty.convert_to_c(store, payload, ctypes.pointer(raw))
            return (name, ffi.wasmtime_component_val_new(byref(raw)))
        raise ValueError('value not valid for this variant')

    def _lift(self, tag: str, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> Any:
        tagged = self._tagged()
        for name, ty in self._cases():
            if name != tag:
                continue
            if ty is None:
                if tagged:
                    return Variant(tag)
                return None
            payload = ty.convert_from_c(ptr.contents)
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_BOOL
            ptr.contents.of.boolean = False
            ffi.wasmtime_component_val_delete(ptr)
            if tagged:
                return Variant(tag, payload)
            return payload
        raise ValueError(f"tag {tag} not found in variant cases")


class VariantType(Managed["ctypes._Pointer[ffi.wasmtime_component_variant_type_t]"], ValType, VariantLikeType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_variant_type_t]") -> None:
        ffi.wasmtime_component_variant_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_variant_type_t]") -> "VariantType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_variant_type_t)):
            raise TypeError("wrong pointer type")
        ty: "VariantType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def cases(self) -> List[Tuple[str, Optional['ValType']]]:
        """
        Returns the cases of this variant type.
        """
        n = ffi.wasmtime_component_variant_type_case_count(self.ptr())
        items = []
        for i in range(n):
            name_ptr = ctypes.POINTER(ctypes.c_char)()
            name_len = ctypes.c_size_t()
            has_payload = ctypes.c_bool()
            valtype_ptr = ffi.wasmtime_component_valtype_t()
            found = ffi.wasmtime_component_variant_type_case_nth(self.ptr(),
                                                                 i,
                                                                 byref(name_ptr),
                                                                 byref(name_len),
                                                                 byref(has_payload),
                                                                 byref(valtype_ptr))
            assert(found)
            name = ctypes.string_at(name_ptr, name_len.value).decode('utf-8')
            if has_payload.value:
                valtype = valtype_from_ptr(valtype_ptr)
            else:
                valtype = None
            items.append((name, valtype))
        return items

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, VariantType):
            return False
        return ffi.wasmtime_component_variant_type_equal(self.ptr(), other.ptr())

    def _cases(self) -> List[Tuple[str, Optional['ValType']]]:
        return self.cases

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        name, raw = self._lower(store, val)
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_VARIANT
        ptr.contents.of.variant.discriminant = ffi.str_to_capi(name)
        ptr.contents.of.variant.val = raw

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_VARIANT.value)
        tag = ffi.to_str(c.of.variant.discriminant)
        ffi.wasm_byte_vec_delete(byref(c.of.variant.discriminant))
        return self._lift(tag, c.of.variant.val)


class EnumType(Managed["ctypes._Pointer[ffi.wasmtime_component_enum_type_t]"], ValType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_enum_type_t]") -> None:
        ffi.wasmtime_component_enum_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_enum_type_t]") -> "EnumType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_enum_type_t)):
            raise TypeError("wrong pointer type")
        ty: "EnumType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def names_count(self) -> int:
        """
        Returns the numter of names of this enum type.
        """
        return ffi.wasmtime_component_enum_type_names_count(self.ptr())

    def name(self, n: int) -> Optional[str]:
        """
        Returns the nth name of this enum type.
        """
        name_ptr = ctypes.POINTER(ctypes.c_char)()
        name_len = ctypes.c_size_t()
        found = ffi.wasmtime_component_enum_type_names_nth(self.ptr(),
                                                           n,
                                                           byref(name_ptr),
                                                           byref(name_len))
        if found:
            return ctypes.string_at(name_ptr, name_len.value).decode('utf-8')
        return None

    @property
    def names(self) -> List[str]:
        """
        Returns the names of this enum type.
        """
        items = []
        for i in range(self.names_count):
            name = self.name(i)
            assert(name is not None)
            items.append(name)
        return items

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, EnumType):
            return False
        return ffi.wasmtime_component_enum_type_equal(self.ptr(), other.ptr())

    def add_classes(self, s: Set[type]) -> None:
        s.add(str)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if isinstance(val, str):
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_ENUM
            ptr.contents.of.enumeration = ffi.str_to_capi(val)
        else:
            raise TypeError("expected str type")

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_ENUM.value)
        ret = ffi.to_str(c.of.enumeration)
        ffi.wasm_byte_vec_delete(byref(c.of.enumeration))
        return ret


class OptionType(Managed["ctypes._Pointer[ffi.wasmtime_component_option_type_t]"], ValType, VariantLikeType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_option_type_t]") -> None:
        ffi.wasmtime_component_option_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_option_type_t]") -> "OptionType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_option_type_t)):
            raise TypeError("wrong pointer type")
        ty: "OptionType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def payload(self) -> 'ValType':
        """
        Returns the payload type of this option type.
        """
        valtype_ptr = ffi.wasmtime_component_valtype_t()
        ffi.wasmtime_component_option_type_ty(self.ptr(), byref(valtype_ptr))
        return valtype_from_ptr(valtype_ptr)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OptionType):
            return False
        return ffi.wasmtime_component_option_type_equal(self.ptr(), other.ptr())

    def _cases(self) -> List[Tuple[str, Optional['ValType']]]:
        return [('none', None), ('some', self.payload)]

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        _, raw = self._lower(store, val)
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_OPTION
        ptr.contents.of.option = raw

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_OPTION.value)
        if c.of.option:
            tag = 'some'
        else:
            tag = 'none'
        return self._lift(tag, c.of.option)



class ResultType(Managed["ctypes._Pointer[ffi.wasmtime_component_result_type_t]"], ValType, VariantLikeType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_result_type_t]") -> None:
        ffi.wasmtime_component_result_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_result_type_t]") -> "ResultType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_result_type_t)):
            raise TypeError("wrong pointer type")
        ty: "ResultType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def ok(self) -> Optional['ValType']:
        """
        Returns the ok type of this result type, if any.
        """
        valtype_ptr = ffi.wasmtime_component_valtype_t()
        has_ok = ffi.wasmtime_component_result_type_ok(self.ptr(), byref(valtype_ptr))
        if not has_ok:
            return None
        return valtype_from_ptr(valtype_ptr)

    @property
    def err(self) -> Optional['ValType']:
        """
        Returns the err type of this result type, if any.
        """
        valtype_ptr = ffi.wasmtime_component_valtype_t()
        has_err = ffi.wasmtime_component_result_type_err(self.ptr(), byref(valtype_ptr))
        if not has_err:
            return None
        return valtype_from_ptr(valtype_ptr)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ResultType):
            return False
        return ffi.wasmtime_component_result_type_equal(self.ptr(), other.ptr())

    def _cases(self) -> List[Tuple[str, Optional['ValType']]]:
        return [('ok', self.ok), ('err', self.err)]

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        name, raw = self._lower(store, val)
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_RESULT
        ptr.contents.of.result.is_ok = name == 'ok'
        ptr.contents.of.result.val = raw

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_RESULT.value)
        if c.of.result.is_ok:
            tag = 'ok'
        else:
            tag = 'err'
        return self._lift(tag, c.of.result.val)


class FlagsType(Managed["ctypes._Pointer[ffi.wasmtime_component_flags_type_t]"], ValType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_flags_type_t]") -> None:
        ffi.wasmtime_component_flags_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_flags_type_t]") -> "FlagsType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_flags_type_t)):
            raise TypeError("wrong pointer type")
        ty: "FlagsType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def names(self) -> List[str]:
        """
        Returns the names of this flags type.
        """
        n = ffi.wasmtime_component_flags_type_names_count(self.ptr())
        items = []
        for i in range(n):
            name_ptr = ctypes.POINTER(ctypes.c_char)()
            name_len = ctypes.c_size_t()
            found = ffi.wasmtime_component_flags_type_names_nth(self.ptr(),
                                                                i,
                                                                byref(name_ptr),
                                                                byref(name_len))
            assert(found)
            name = ctypes.string_at(name_ptr, name_len.value).decode('utf-8')
            items.append(name)
        return items

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FlagsType):
            return False
        return ffi.wasmtime_component_flags_type_equal(self.ptr(), other.ptr())

    def add_classes(self, s: Set[type]) -> None:
        s.add(set)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if not isinstance(val, set):
            raise TypeError("expected set type for Flags type")
        for s in val:
            if not isinstance(s, str):
                raise TypeError("expected set of strings for Flags type")

        raw = ffi.wasmtime_component_valflags_t()
        ffi.wasmtime_component_valflags_new_uninit(raw, len(val))
        for i, s in enumerate(val):
            raw.data[i] = ffi.str_to_capi(s)
        ptr.contents.kind = ffi.WASMTIME_COMPONENT_FLAGS
        ptr.contents.of.flags = raw

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_FLAGS.value)
        result = set()
        for i in range(c.of.flags.size):
            s = ffi.to_str(c.of.flags.data[i])
            result.add(s)
        ffi.wasmtime_component_valflags_delete(byref(c.of.flags))
        return result


class FutureType(Managed["ctypes._Pointer[ffi.wasmtime_component_future_type_t]"], ValType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_future_type_t]") -> None:
        ffi.wasmtime_component_future_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_future_type_t]") -> "FutureType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_future_type_t)):
            raise TypeError("wrong pointer type")
        ty: "FutureType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def payload(self) -> 'ValType':
        """
        Returns the payload type of this future type.
        """
        valtype_ptr = ffi.wasmtime_component_valtype_t()
        ffi.wasmtime_component_future_type_ty(self.ptr(), byref(valtype_ptr))
        return valtype_from_ptr(valtype_ptr)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FutureType):
            return False
        return ffi.wasmtime_component_future_type_equal(self.ptr(), other.ptr())

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        raise NotImplementedError("Future conversion not implemented yet")

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        raise NotImplementedError("conversion not implemented yet")


class StreamType(Managed["ctypes._Pointer[ffi.wasmtime_component_stream_type_t]"], ValType):
    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_stream_type_t]") -> None:
        ffi.wasmtime_component_stream_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_stream_type_t]") -> "StreamType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_stream_type_t)):
            raise TypeError("wrong pointer type")
        ty: "StreamType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def payload(self) -> 'ValType':
        """
        Returns the payload type of this stream type.
        """
        valtype_ptr = ffi.wasmtime_component_valtype_t()
        ffi.wasmtime_component_stream_type_ty(self.ptr(), byref(valtype_ptr))
        return valtype_from_ptr(valtype_ptr)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StreamType):
            return False
        return ffi.wasmtime_component_stream_type_equal(self.ptr(), other.ptr())

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        raise NotImplementedError("Stream conversion not implemented yet")

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        raise NotImplementedError("conversion not implemented yet")


@dataclass
class OwnType(ValType):
    ty: ResourceType

    def add_classes(self, s: Set[type]) -> None:
        s.add(ResourceAny)
        s.add(ResourceHost)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if isinstance(val, ResourceAny):
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_RESOURCE
            ptr.contents.of.resource = val._consume()
        elif isinstance(val, ResourceHost):
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_RESOURCE
            ptr.contents.of.resource = val.to_any(store)._consume()
        else:
            raise TypeError("expected ResourceAny or ResourceHost for Own type")

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_RESOURCE.value)
        return ResourceAny._from_ptr(ffi.take_pointer(c.of, 'resource'))


@dataclass
class BorrowType(ValType):
    ty: ResourceType

    def add_classes(self, s: Set[type]) -> None:
        s.add(ResourceAny)
        s.add(ResourceHost)

    def convert_to_c(self, store: Storelike, val: Any, ptr: 'ctypes._Pointer[ffi.wasmtime_component_val_t]') -> None:
        if isinstance(val, ResourceAny):
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_RESOURCE
            ptr.contents.of.resource = ffi.wasmtime_component_resource_any_clone(val.ptr())
        elif isinstance(val, ResourceHost):
            ptr.contents.kind = ffi.WASMTIME_COMPONENT_RESOURCE
            ptr.contents.of.resource = val.to_any(store)._consume()
        else:
            raise TypeError("expected ResourceAny or ResourceHost for Own type")

    def convert_from_c(self, c: 'ffi.wasmtime_component_val_t') -> Any:
        assert(c.kind == ffi.WASMTIME_COMPONENT_RESOURCE.value)
        return ResourceAny._from_ptr(ffi.take_pointer(c.of, 'resource'))


ComponentItem = Union[
    ComponentType,
    ModuleType,
    ComponentInstanceType,
    ResourceType,
    ValType,
    FuncType,
]


def component_item_from_ptr(ptr: ffi.wasmtime_component_item_t) -> ComponentItem:
    if ptr.kind == ffi.WASMTIME_COMPONENT_ITEM_COMPONENT.value:
        return ComponentType._from_ptr(ptr.of.component)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_ITEM_MODULE.value:
        return ModuleType._from_ptr(ptr.of.module)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_ITEM_COMPONENT_INSTANCE.value:
        return ComponentInstanceType._from_ptr(ptr.of.component_instance)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_ITEM_RESOURCE.value:
        return ResourceType._from_ptr(ptr.of.resource)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_ITEM_COMPONENT_FUNC.value:
        return FuncType._from_ptr(ptr.of.component_func)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_ITEM_TYPE.value:
        return valtype_from_ptr(ptr.of.type)
    else:
        ffi.wasmtime_component_item_delete(byref(ptr))
        raise TypeError("unknown component item kind")


def valtype_from_ptr(ptr: ffi.wasmtime_component_valtype_t) -> ValType:
    if ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_BOOL.value:
        return Bool()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_S8.value:
        return S8()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_S16.value:
        return S16()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_S32.value:
        return S32()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_S64.value:
        return S64()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_U8.value:
        return U8()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_U16.value:
        return U16()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_U32.value:
        return U32()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_U64.value:
        return U64()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_F32.value:
        return F32()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_F64.value:
        return F64()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_CHAR.value:
        return Char()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_STRING.value:
        return String()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_ERROR_CONTEXT.value:
        return ErrorContext()
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_LIST.value:
        return ListType._from_ptr(ptr.of.list)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_RECORD.value:
        return RecordType._from_ptr(ptr.of.record)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_TUPLE.value:
        return TupleType._from_ptr(ptr.of.tuple)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_VARIANT.value:
        return VariantType._from_ptr(ptr.of.variant)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_ENUM.value:
        return EnumType._from_ptr(ptr.of.enum_)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_OPTION.value:
        return OptionType._from_ptr(ptr.of.option)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_RESULT.value:
        return ResultType._from_ptr(ptr.of.result)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_FLAGS.value:
        return FlagsType._from_ptr(ptr.of.flags)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_FUTURE.value:
        return FutureType._from_ptr(ptr.of.future)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_STREAM.value:
        return StreamType._from_ptr(ptr.of.stream)
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_OWN.value:
        return OwnType(ResourceType._from_ptr(ptr.of.own))
    elif ptr.kind == ffi.WASMTIME_COMPONENT_VALTYPE_BORROW.value:
        return BorrowType(ResourceType._from_ptr(ptr.of.borrow))
    else:
        ffi.wasmtime_component_valtype_delete(byref(ptr))
        raise TypeError("unknown component value type kind")



import unittest
from typing import cast
import gc

from wasmtime import *


def ref_types_store():
    config = Config()
    config.wasm_reference_types = True
    engine = Engine(config)
    return Store(engine)


def compile_and_instantiate(wat):
    store = ref_types_store()
    module = Module(store.engine, wat)
    return (Instance(store, module, []), store)


class SetHitOnDrop:
    def __init__(self, obj):
        obj['hit'] = False
        self.obj = obj

    def __del__(self):
        self.obj['hit'] = True


class TestExternRef(unittest.TestCase):
    def test_smoke(self):
        (instance, store) = compile_and_instantiate(
            """
            (module
                 (func (export "f") (param externref) (result externref)
                     local.get 0
                 )
                 (func (export "null_externref") (result externref)
                     ref.null extern
                 )
            )
            """
        )

        null_externref = instance.exports(store).get("null_externref")
        self.assertEqual(null_externref(store), None)

        f = instance.exports(store).get("f")
        externs = [42, True, False, None, "Hello", {"x": 1}, [12, 13, 14], Config()]

        for extern in externs:
            # We can create an externref for the given extern data.
            ref = Val.externref(extern)

            # And the externref's value is our extern data.
            self.assertEqual(ref.value, extern)

            # And we can round trip the externref through Wasm and still get our
            # extern data.
            result = f(store, ref)
            self.assertEqual(result, extern)

    def test_int_to_externref(self):
        wat = """
            (module
                 (import "env" "int_to_ref" (func $int_to_ref (param $a externref) (result externref)))
                 (export "test" (func $int_to_ref))
            )
            """
        config = Config()
        config.wasm_reference_types = True
        engine = Engine(config)
        store = Store(engine)
        module = Module(store.engine, wat)
        linker = Linker(engine)
        ftype = FuncType([ValType.externref()], [ValType.externref()])
        linker.define_func("env", "int_to_ref", ftype, lambda x: x)
        instance = linker.instantiate(store, module)
        f: Func
        f = cast(Func, instance.exports(store).get("test"))
        f(store, 5)
        f = cast(Func, instance.exports(store).get("test"))
        f(store, 5.7)

    def test_externref_tables(self):
        store = ref_types_store()
        ty = TableType(ValType.externref(), Limits(10, None))
        table = Table(store, ty, "init")

        for i in range(0, 10):
            self.assertEqual(table.get(store, i), "init")

        table.grow(store, 2, "grown")

        for i in range(0, 10):
            self.assertEqual(table.get(store, i), "init")
        for i in range(10, 12):
            self.assertEqual(table.get(store, i), "grown")

        table.set(store, 7, "lucky")

        for i in range(0, 7):
            self.assertEqual(table.get(store, i), "init")
        self.assertEqual(table.get(store, 7), "lucky")
        for i in range(8, 10):
            self.assertEqual(table.get(store, i), "init")
        for i in range(10, 12):
            self.assertEqual(table.get(store, i), "grown")

    def test_externref_in_global(self):
        store = ref_types_store()
        ty = GlobalType(ValType.externref(), True)
        g = Global(store, ty, Val.externref("hello"))
        self.assertEqual(g.value(store), "hello")
        g.set_value(store, "goodbye")
        self.assertEqual(g.value(store), "goodbye")

    def test_dtor_global(self):
        obj = {}  # type: ignore
        store = ref_types_store()
        ty = GlobalType(ValType.externref(), True)
        g = Global(store, ty, Val.externref(SetHitOnDrop(obj)))
        assert(not obj['hit'])
        g.set_value(store, None)
        store.gc()
        gc.collect()
        assert(obj['hit'])

    def test_dtor_func(self):
        (instance, store) = compile_and_instantiate(
            """
            (module
                 (func (export "f") (param externref))
            )
            """
        )

        f = instance.exports(store).get("f")
        obj = {}  # type: ignore
        f(store, SetHitOnDrop(obj))
        store.gc()
        gc.collect()
        assert(obj['hit'])


class TestFuncRef(unittest.TestCase):
    def test_smoke(self):
        (instance, store) = compile_and_instantiate(
            """
            (module
                 (func (export \"f\") (param funcref) (result funcref)
                     local.get 0
                 )
                 (func (export "null_funcref") (result funcref)
                     ref.null func
                 )
            )
            """
        )

        null_funcref = instance.exports(store).get("null_funcref")
        self.assertEqual(null_funcref(store), None)

        f = instance.exports(store).get("f")

        ty = FuncType([], [ValType.i32()])
        g = Func(store, ty, lambda: 42)

        # We can create a funcref.
        ref_g_val = Val.funcref(g)

        # And the funcref's points to `g`.
        g2 = ref_g_val.as_funcref()
        if isinstance(g2, Func):
            self.assertEqual(g2(store), 42)
        else:
            self.fail("g2 is not a funcref: g2 = %r" % g2)

        # And we can round trip the funcref through Wasm.
        g3 = f(store, ref_g_val)
        if isinstance(g3, Func):
            self.assertEqual(g3(store), 42)
        else:
            self.fail("g3 is not a funcref: g3 = %r" % g3)



from abc import abstractmethod
from typing import TypeVar, Generic, Any, Optional

T = TypeVar('T')

class Managed(Generic[T]):
    """
    Abstract base class for types which contain an owned pointer in the C FFI
    layer.

    Not exported directly from this package.
    """
    __ptr: Optional[T]

    @abstractmethod
    def _delete(self, ptr: T) -> None:
        """
        Runs the FFI destructor for the `ptr` specified.

        Must be implemented by classes that inherit from this class.
        """
        pass

    def _set_ptr(self, ptr: T) -> None:
        if hasattr(self, '__ptr'):
            raise ValueError('already initialized')
        self.__ptr = ptr

    def close(self) -> None:
        """
        Closes this object, or deallocates it. Further usage of this object
        will raise a `ValueError`.
        """

        # Get the pointer's value but don't worry if it was never set.
        try:
            ptr = self.__ptr
        except AttributeError:
            return

        # If it wasn't previously deallocated then do so here, otherwise ignore
        # this repeated call to `close`
        if ptr is not None:
            self._delete(ptr)
            self.__ptr = None

    def ptr(self) -> T:
        """
        Returns the underlying pointer for this FFI object, or a `ValueError`
        if it's already been closed.
        """

        # Fail with a `ValueError` if the pointer was never set
        try:
            ptr = self.__ptr
        except AttributeError:
            raise ValueError('never initialized')

        # Check to see if this object is already deallocated, and if so raise
        # a specific exception
        if ptr is not None:
            return ptr
        else:
            raise ValueError('already closed')

    def _consume(self) -> T:
        """
        Internal method to take ownership of the internal pointer without
        destroying it.
        """
        ret = self.ptr()
        self.__ptr = None
        return ret

    def __enter__(self) -> Any:
        """
        Entry part of the contextlib protocol to enable using this object with
        `with`.

        Returns `self` to bind to use within a `with` block.
        """
        return self

    def __exit__(self, *exc: Any) -> None:
        """
        Exit part of the contextlib protocol to call `close`.
        """
        self.close()

    def __del__(self) -> None:
        """
        Automatic destruction of the internal FFI object if it's still alive by
        this point.
        """
        self.close()



import unittest

from wasmtime import *


class TestSharedMemory(unittest.TestCase):
    def test_new(self):
        engine = Store().engine
        memory_type = MemoryType(Limits(1, 2), is_64=False, shared=True)
        assert(not memory_type.is_64)
        shared_memory = SharedMemory(engine, memory_type)
        with self.assertRaises(TypeError):
            shared_memory.grow('')  # type: ignore
        with self.assertRaises(WasmtimeError):
            shared_memory.grow(-1)
        self.assertEqual(shared_memory.data_ptr()[0], 0)
        self.assertEqual(shared_memory.data_len(), 65536)
        self.assertTrue(isinstance(shared_memory.type(), MemoryType))

    def test_grow(self):
        engine = Store().engine
        memory_type = MemoryType(Limits(1, 2), shared=True)
        shared_memory = SharedMemory(engine, memory_type)
        assert(shared_memory.grow(1) == 1)
        assert(shared_memory.grow(0) == 2)
        with self.assertRaises(WasmtimeError):
            shared_memory.grow(1)

    def test_errors(self):
        engine = Store().engine
        ty = MemoryType(Limits(1, 2), shared=True)
        with self.assertRaises(AttributeError):
            SharedMemory(1, ty)  # type: ignore
        with self.assertRaises(AttributeError):
            SharedMemory(engine, 1)  # type: ignore

    def test_shared_memory_type_fails_if_no_max_specified(self):
        with self.assertRaises(WasmtimeError):
            # Shared memories must have a max size
            MemoryType(Limits(0x100000000, None), shared=True)

    def test_shared_memory_type_works_if_min_max_i64_is_set(self):
        ty = MemoryType(Limits(0x100000000, 0x100000000), is_64=True, shared=True)
        assert(ty.limits.min == 0x100000000)
        assert(ty.limits.max == 0x100000000)
        assert(ty.is_64)
        
    def test_shared_memory_type_fails_if_is_too_large(self):
        with self.assertRaises(WasmtimeError):
            MemoryType(Limits(1, 0x100000000), shared=True)

    def test_memory_type_has_to_be_shared(self):
        engine = Store().engine
        ty = MemoryType(Limits(1, 2))
        with self.assertRaises(WasmtimeError):
            SharedMemory(engine, ty)



import unittest

from wasmtime import *


class TestStore(unittest.TestCase):
    def test_smoke(self):
        Store()
        Store(Engine())

        with Store() as store:
            pass

        store = Store()
        store.close()
        store.close()
        with self.assertRaises(ValueError):
            store.set_epoch_deadline(1)

    def test_errors(self):
        with self.assertRaises(TypeError):
            Store(3)  # type: ignore

    def test_interrupt_handle(self):
        config = Config()
        config.epoch_interruption = True
        engine = Engine(config)
        engine.increment_epoch()
        store = Store(engine)
        store.set_epoch_deadline(1)

    def test_interrupt_wasm(self):
        config = Config()
        config.epoch_interruption = True
        engine = Engine(config)
        store = Store(engine)
        store.set_epoch_deadline(1)

        module = Module(store.engine, """
            (import "" "hit" (func $hit))
            (import "" "interrupt" (func $interrupt))
            (func $start
                call $hit
                call $interrupt
                (loop br 0))
            (start $start)
        """)
        interrupt = Func(store, FuncType([], []), lambda: engine.increment_epoch())

        was_hit = False

        def hit_callback():
            nonlocal was_hit
            was_hit = True
        hit = Func(store, FuncType([], []), hit_callback)

        with self.assertRaises(Trap):
            Instance(store, module, [hit, interrupt])
        self.assertTrue(was_hit)

    def test_fuel(self):
        store = Store()

        with self.assertRaises(WasmtimeError):
            store.set_fuel(1)
        with self.assertRaises(WasmtimeError):
            store.get_fuel()

        config = Config()
        config.consume_fuel = True
        store = Store(Engine(config))
        store.set_fuel(1)
        assert(store.get_fuel() == 1)
        store.set_fuel(2)
        assert(store.get_fuel() == 2)
        store.set_fuel(0)
        assert(store.get_fuel() == 0)

    def test_limits(self):
        store = Store()
        Memory(store, MemoryType(Limits(1, None)))

        store = Store()
        store.set_limits(memory_size=0)
        with self.assertRaises(WasmtimeError):
            Memory(store, MemoryType(Limits(1, None)))
        store.set_limits(memory_size=100000)
        Memory(store, MemoryType(Limits(1, None)))

        store = Store()
        store.set_limits(table_elements=1)
        Table(store, TableType(ValType.funcref(), Limits(1, None)), None)
        with self.assertRaises(WasmtimeError):
            Table(store, TableType(ValType.funcref(), Limits(2, None)), None)

        store = Store()
        store.set_limits(memory_size=200000)
        mem = Memory(store, MemoryType(Limits(1, None)))
        mem.grow(store, 1)
        with self.assertRaises(WasmtimeError):
            mem.grow(store, 100)



import unittest

from wasmtime import *


class TestFunc(unittest.TestCase):
    def test_smoke(self):
        store = Store()
        ty = FuncType([], [])
        func = Func(store, ty, lambda: None)
        func(store)
        self.assertTrue(isinstance(func.type(store), FuncType))

    def test_add(self):
        store = Store()
        ty = FuncType([ValType.i32(), ValType.i32()], [ValType.i32()])
        func = Func(store, ty, lambda a, b: a + b)
        self.assertEqual(func(store, 1, 2), 3)

    def test_calls(self):
        store = Store()
        ty = FuncType([ValType.i32()], [])
        func = Func(store, ty, lambda a: self.assertEqual(a, 1))
        func(store, 1)

        ty = FuncType([ValType.i64()], [])
        func = Func(store, ty, lambda a: self.assertEqual(a, 2))
        func(store, Val.i64(2))

        ty = FuncType([ValType.f32()], [])
        func = Func(store, ty, lambda a: self.assertEqual(a, 3.0))
        func(store, 3.0)

        ty = FuncType([ValType.f64()], [])
        func = Func(store, ty, lambda a: self.assertEqual(a, 4.0))
        func(store, 4.0)

    def test_multi_return(self):
        store = Store()
        ty = FuncType([], [ValType.i32(), ValType.i32()])
        func = Func(store, ty, lambda: [1, 2])
        self.assertEqual(func(store), [1, 2])

    def test_errors(self):
        store = Store()
        ty = FuncType([], [])
        with self.assertRaises(TypeError):
            Func(1, ty, lambda: None)  # type: ignore
        with self.assertRaises(TypeError):
            Func(store, 1, lambda: None)  # type: ignore
        func = Func(store, ty, lambda: None)
        with self.assertRaises(WasmtimeError):
            func(store, 2)

        ty = FuncType([ValType.i32()], [])
        func = Func(store, ty, lambda: None)
        with self.assertRaises(TypeError):
            func(store, 3.0)
        with self.assertRaises(TypeError):
            func(store, Val.i64(3))
        ty = FuncType([ValType.i32()], [])

        func = Func(store, ty, lambda x: x)
        with self.assertRaises(WasmtimeError, msg="produced results"):
            func(store, 1)

    def test_produce_wrong(self):
        store = Store()
        ty = FuncType([], [ValType.i32(), ValType.i32()])
        func = Func(store, ty, lambda: 1)
        with self.assertRaises(TypeError, msg="has no len"):
            func(store)
        func = Func(store, ty, lambda: [1, 2, 3])
        with self.assertRaises(WasmtimeError, msg="wrong number of results"):
            func(store)

    def test_host_exception(self):
        store = Store()
        ty = FuncType([], [])

        def do_raise():
            raise Exception("hello")

        func = Func(store, ty, do_raise)
        with self.assertRaises(Exception, msg="hello"):
            func(store)

    def test_type(self):
        store = Store()
        i32 = ValType.i32()
        i64 = ValType.i64()
        f32 = ValType.f32()
        f64 = ValType.f64()
        ty = FuncType([i32, i64, f32, f64], [f64, f32, i64, i32])

        def rev(*args):
            ret = list(args)
            ret.reverse()
            return ret

        func = Func(store, ty, rev)
        self.assertEqual(func(store, 1, 2, 3.0, 4.0), [4.0, 3.0, 2, 1])

    def test_access_caller(self):
        # Test that we get *something*
        store = Store()

        def runtest(caller):
            self.assertEqual(caller.get(''), None)
            self.assertEqual(caller.get('x'), None)
            self.assertEqual(caller.get('y'), None)

        Func(store, FuncType([], []), runtest, access_caller=True)(store)

        hit = {}

        # Test that `Caller` works and that it's invalidated
        def runtest2(caller):
            hit['yes'] = True
            hit['caller'] = caller

            self.assertTrue(caller.get('bar') is None)
            mem = caller.get('foo')
            self.assertTrue(isinstance(mem, Memory))

            self.assertEqual(mem.data_ptr(caller)[0], ord('f'))
            self.assertEqual(mem.data_ptr(caller)[1], ord('o'))
            self.assertEqual(mem.data_ptr(caller)[2], ord('o'))
            self.assertEqual(mem.data_ptr(caller)[3], 0)

        module = Module(store.engine, """
            (module
                (import "" "" (func))
                (memory (export "foo") 1)
                (start 0)
                (data (i32.const 0) "foo")
            )
        """)
        func = Func(store, FuncType([], []), runtest2, access_caller=True)
        Instance(store, module, [func])
        self.assertTrue(hit['yes'])
        self.assertTrue(hit['caller'].get('foo') is None)  # type: ignore

        # Test that `Caller` is invalidated even on exceptions
        hit2 = {}

        def runtest3(caller):
            hit2['caller'] = caller
            self.assertTrue(caller['foo'] is not None)
            raise WasmtimeError('foo')

        func = Func(store, FuncType([], []), runtest3, access_caller=True)
        with self.assertRaises(WasmtimeError, msg='foo'):
            Instance(store, module, [func])
        self.assertTrue(hit2['caller'].get('foo') is None)



import ctypes

from . import _ffi as ffi
from wasmtime import GlobalType, Val, WasmtimeError
from typing import Any
from ._store import Storelike


class Global:
    _global: ffi.wasmtime_global_t

    def __init__(self, store: Storelike, ty: GlobalType, val: Any):
        if not isinstance(ty, GlobalType):
            raise TypeError("expected a GlobalType")
        val = Val._convert_to_raw(store, ty.content, val)
        global_ = ffi.wasmtime_global_t()
        error = ffi.wasmtime_global_new(
            store._context(),
            ty.ptr(),
            ctypes.byref(val),
            ctypes.byref(global_))
        ffi.wasmtime_val_unroot(ctypes.byref(val))
        if error:
            raise WasmtimeError._from_ptr(error)
        self._global = global_

    @classmethod
    def _from_raw(cls, global_: ffi.wasmtime_global_t) -> "Global":
        ty: "Global" = cls.__new__(cls)
        ty._global = global_
        return ty

    def type(self, store: Storelike) -> GlobalType:
        """
        Gets the type of this global as a `GlobalType`
        """

        ptr = ffi.wasmtime_global_type(store._context(), ctypes.byref(self._global))
        return GlobalType._from_ptr(ptr, None)

    def value(self, store: Storelike) -> Any:
        """
        Gets the current value of this global

        Returns a native python type
        """
        raw = ffi.wasmtime_val_t()
        ffi.wasmtime_global_get(store._context(), ctypes.byref(self._global), ctypes.byref(raw))
        val = Val._from_raw(store, raw)
        if val.value is not None:
            return val.value
        else:
            return val

    def set_value(self, store: Storelike, val: Any) -> None:
        """
        Sets the value of this global to a new value
        """
        val = Val._convert_to_raw(store, self.type(store).content, val)
        error = ffi.wasmtime_global_set(store._context(), ctypes.byref(self._global), ctypes.byref(val))
        ffi.wasmtime_val_unroot(ctypes.byref(val))
        if error:
            raise WasmtimeError._from_ptr(error)

    def _as_extern(self) -> ffi.wasmtime_extern_t:
        union = ffi.wasmtime_extern_union(global_=self._global)
        return ffi.wasmtime_extern_t(ffi.WASMTIME_EXTERN_GLOBAL, union)



from . import _ffi as ffi
from ctypes import byref, c_uint64, cast, c_void_p, CFUNCTYPE
import ctypes
from wasmtime import Engine, WasmtimeError, Managed
from . import _value as value
import typing

if typing.TYPE_CHECKING:
    from ._wasi import WasiConfig


class Store(Managed["ctypes._Pointer[ffi.wasmtime_store_t]"]):
    __context: "typing.Optional[ctypes._Pointer[ffi.wasmtime_context_t]]"

    def __init__(self, engine: typing.Optional[Engine] = None, data: typing.Optional[typing.Any] = None):

        if engine is None:
            engine = Engine()
        elif not isinstance(engine, Engine):
            raise TypeError("expected an Engine")
        data_id = None
        finalize = cast(0, CFUNCTYPE(None, c_void_p))
        if data:
            data_id = value._intern(data)
            finalize = value._externref_finalizer
        self._set_ptr(ffi.wasmtime_store_new(engine.ptr(), data_id, finalize))
        self.__context = ffi.wasmtime_store_context(self.ptr())
        self.engine = engine

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_store_t]") -> None:
        ffi.wasmtime_store_delete(ptr)
        self.__context = None

    def _context(self) -> "ctypes._Pointer[ffi.wasmtime_context_t]":
        if self.__context is None:
            raise ValueError('already closed')
        return self.__context

    def data(self) -> typing.Optional[typing.Any]:
        """
        TODO
        """
        data = ffi.wasmtime_context_get_data(self._context())
        if data:
            # FIXME https://github.com/bytecodealliance/wasmtime-py/issues/303
            return value._unintern(data)  # type: ignore
        else:
            return None

    def gc(self) -> None:
        """
        Runs a GC over `externref` values that have been passed into this Store,
        cleaning out anything that is no longer referenced.

        This is not required to be run manually, but can be done so if you'd
        like more precise control over when unreferenced `externref` values are
        deallocated.
        """
        ffi.wasmtime_context_gc(self._context())

    def set_fuel(self, fuel: int) -> None:
        """
        Sets the amount of fuel in this store to `fuel`.

        This is only relevant when `Config.consume_fuel` is configured.

        This is a required call to ensure that the store has fuel to
        execute WebAssembly since otherwise stores start with zero fuel.

        Raises a `WasmtimeError` if this store's configuration is not configured
        to consume fuel.
        """
        err = ffi.wasmtime_context_set_fuel(self._context(), fuel)
        if err:
            raise WasmtimeError._from_ptr(err)

    def get_fuel(self) -> int:
        """
        Returns the amount of fuel left in the store.

        This is only relevant when `Config.consume_fuel` is configured.

        Raises a `WasmtimeError` if this store's configuration is not configured
        to consume fuel or if the store doesn't have enough fuel remaining.
        """
        remaining = c_uint64(0)
        err = ffi.wasmtime_context_get_fuel(self._context(), byref(remaining))
        if err:
            raise WasmtimeError._from_ptr(err)
        return remaining.value

    def set_wasi(self, wasi: "WasiConfig") -> None:
        """
        TODO
        """
        error = ffi.wasmtime_context_set_wasi(self._context(), wasi._consume())
        if error:
            raise WasmtimeError._from_ptr(error)

    def set_epoch_deadline(self, ticks_after_current: int) -> None:
        """
        Configures the relative epoch deadline, after the current engine's
        epoch, after which WebAssembly code will trap.
        """
        ffi.wasmtime_context_set_epoch_deadline(self._context(), ticks_after_current)

    def set_limits(self,
                   memory_size: int = -1,
                   table_elements: int = -1,
                   instances: int = -1,
                   tables: int = -1,
                   memories: int = -1) -> None:
        """
        Configures the limits of various items within this store.

        * `memory_size` - the maximum size, in bytes, that linear memory is
          allowed to consume within this store. Setting this to a lower value
          will cause instantiation to fail if a module needs more memory.
          Additionally the `memory.grow` instruction will return -1 once this
          threshold is reached.

        * `table_elements` - the maximum number of elements that can be stored
          within tables in this store. Currently each table element takes 8
          bytes.

        * `instances` - the maximum number of WebAssembly instances that can
          be created.

        * `tables` - the maximum number of WebAssembly tables that can
          be created.

        * `memories` - the maximum number of WebAssembly linear memories that
          can be created.

        If any limit is negative then the limit will not be set as a part of
        this invocation and it will be ignored.
        """
        ffi.wasmtime_store_limiter(self.ptr(), memory_size, table_elements, instances, tables, memories)


class StoreContext:
    __ptr: typing.Optional["ctypes._Pointer[ffi.wasmtime_context_t]"]

    def __init__(self, ptr: "ctypes._Pointer[ffi.wasmtime_context_t]"):
        self.__ptr = ptr

    def _context(self) -> "ctypes._Pointer[ffi.wasmtime_context_t]":
        if self.__ptr is None:
            raise ValueError("caller is no longer valid")
        return self.__ptr

    def _invalidate(self) -> None:
        self.__ptr = None


if typing.TYPE_CHECKING:
    from ._func import Caller


Storelike = typing.Union[Store, "Caller", StoreContext]



from . import _ffi as ffi
from enum import Enum
from ctypes import byref, POINTER
import ctypes
from typing import Optional, Any, List
from wasmtime import Managed


class TrapCode(Enum):
    # The current stack space was exhausted.
    STACK_OVERFLOW = 0
    # An out-of-bounds memory access.
    MEMORY_OUT_OF_BOUNDS = 1
    # A wasm atomic operation was presented with a not-naturally-aligned linear-memory address.
    HEAP_MISALIGNED = 2
    # An out-of-bounds access to a table.
    TABLE_OUT_OF_BOUNDS = 3
    # Indirect call to a null table entry.
    INDIRECT_CALL_TO_NULL = 4
    # Signature mismatch on indirect call.
    BAD_SIGNATURE = 5
    # An integer arithmetic operation caused an overflow.
    INTEGER_OVERFLOW = 6
    # An integer division by zero.
    INTEGER_DIVISION_BY_ZERO = 7
    # Failed float-to-int conversion.
    BAD_CONVERSION_TO_INTEGER = 8
    # Code that was supposed to have been unreachable was reached.
    UNREACHABLE = 9
    # Execution has potentially run too long and may be interrupted.
    INTERRUPT = 10


class Trap(Exception, Managed["ctypes._Pointer[ffi.wasm_trap_t]"]):

    def __init__(self, message: str):
        """
        Creates a new trap with the given `message`
        """

        vec = message.encode('utf-8')
        self._set_ptr(ffi.wasmtime_trap_new(ctypes.create_string_buffer(vec), len(vec)))

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasm_trap_t]") -> None:
        ffi.wasm_trap_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasm_trap_t]") -> "Trap":
        if not isinstance(ptr, POINTER(ffi.wasm_trap_t)):
            raise TypeError("wrong pointer type")
        trap: Trap = cls.__new__(cls)
        trap._set_ptr(ptr)
        return trap

    @property
    def message(self) -> str:
        """
        Returns the message for this trap
        """

        message = ffi.wasm_byte_vec_t()
        ffi.wasm_trap_message(self.ptr(), byref(message))
        # subtract one to chop off the trailing nul byte
        message.size -= 1
        ret = ffi.to_str(message)
        message.size += 1
        ffi.wasm_byte_vec_delete(byref(message))
        return ret

    @property
    def frames(self) -> List["Frame"]:
        frames = FrameList(self)
        ffi.wasm_trap_trace(self.ptr(), byref(frames.vec))
        ret = []
        for i in range(0, frames.vec.size):
            ret.append(Frame._from_ptr(frames.vec.data[i], frames))
        return ret

    @property
    def trap_code(self) -> Optional[TrapCode]:
        """
        Returns an optional `TrapCode` that corresponds to why this trap
        happened.

        Note that `None` may be returned for manually created traps which do
        not have an associated code with them.
        """
        code = ffi.wasmtime_trap_code_t()
        if ffi.wasmtime_trap_code(self.ptr(), byref(code)):
            return TrapCode(code.value)
        return None

    def __str__(self) -> str:
        return self.message


class Frame(Managed["ctypes._Pointer[ffi.wasm_frame_t]"]):
    _owner: Optional[Any]

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasm_frame_t]", owner: Optional[Any]) -> "Frame":
        if not isinstance(ptr, POINTER(ffi.wasm_frame_t)):
            raise TypeError("wrong pointer type")
        ty: "Frame" = cls.__new__(cls)
        ty._set_ptr(ptr)
        ty._owner = owner
        return ty

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasm_frame_t]") -> None:
        if self._owner is None:
            ffi.wasm_frame_delete(ptr)

    @property
    def func_index(self) -> int:
        """
        Returns the function index this frame corresponds to in its wasm module
        """

        return ffi.wasm_frame_func_index(self.ptr())

    @property
    def func_name(self) -> Optional[str]:
        """
        Returns the name of the function this frame corresponds to

        May return `None` if no name can be inferred
        """

        ptr = ffi.wasmtime_frame_func_name(self.ptr())
        if ptr:
            return ffi.to_str(ptr.contents)
        else:
            return None

    @property
    def module_name(self) -> Optional[str]:
        """
        Returns the name of the module this frame corresponds to

        May return `None` if no name can be inferred
        """

        ptr = ffi.wasmtime_frame_module_name(self.ptr())
        if ptr:
            return ffi.to_str(ptr.contents)
        else:
            return None

    @property
    def module_offset(self) -> int:
        """
        Returns the offset of this frame's program counter into the original
        wasm source module.
        """

        return ffi.wasm_frame_module_offset(self.ptr())

    @property
    def func_offset(self) -> int:
        """
        Returns the offset of this frame's program counter into the original
        wasm function.
        """

        return ffi.wasm_frame_func_offset(self.ptr())


class FrameList:
    owner: Any

    def __init__(self, owner: Any) -> None:
        self.vec = ffi.wasm_frame_vec_t(0, None)
        self.owner = owner

    def __del__(self) -> None:
        ffi.wasm_frame_vec_delete(byref(self.vec))



from ctypes import POINTER, Structure, Union, addressof, byref, cdll, string_at
from ctypes import c_double, c_float, c_int32, c_int64, c_uint8
from pathlib import Path
import ctypes
import sys
import platform
import typing

if sys.maxsize <= 2**32:
    raise RuntimeError("wasmtime only works on 64-bit platforms right now")

sys_platform = sys.platform

# For Python versions <=3.12. 3.13+ supports PEP 738 and uses sys.platform
if hasattr(sys, 'getandroidapilevel'):
    sys_platform = 'android'

if sys_platform == 'linux' or sys_platform == 'android':
    libname = '_libwasmtime.so'
elif sys_platform == 'win32':
    libname = '_wasmtime.dll'
elif sys_platform == 'darwin':
    libname = '_libwasmtime.dylib'
else:
    raise RuntimeError("unsupported platform `{}` for wasmtime".format(sys_platform))


machine = platform.machine()
if machine == 'AMD64':
    machine = 'x86_64'
if machine == 'arm64' or machine == 'ARM64':
    machine = 'aarch64'
if machine != 'x86_64' and machine != 'aarch64':
    raise RuntimeError("unsupported architecture for wasmtime: {}".format(machine))

filename = Path(__file__).parent / (sys_platform + '-' + machine) / libname

dll = cdll.LoadLibrary(str(filename))

WASM_I32 = c_uint8(0)
WASM_I64 = c_uint8(1)
WASM_F32 = c_uint8(2)
WASM_F64 = c_uint8(3)
WASM_ANYREF = c_uint8(128)
WASM_FUNCREF = c_uint8(129)
# WASM_V128 = c_uint8(4)

WASMTIME_I32 = c_uint8(0)
WASMTIME_I64 = c_uint8(1)
WASMTIME_F32 = c_uint8(2)
WASMTIME_F64 = c_uint8(3)
WASMTIME_V128 = c_uint8(4)
WASMTIME_FUNCREF = c_uint8(5)
WASMTIME_EXTERNREF = c_uint8(6)

WASM_CONST = c_uint8(0)
WASM_VAR = c_uint8(1)

WASMTIME_EXTERN_FUNC = c_uint8(0)
WASMTIME_EXTERN_GLOBAL = c_uint8(1)
WASMTIME_EXTERN_TABLE = c_uint8(2)
WASMTIME_EXTERN_MEMORY = c_uint8(3)
WASMTIME_EXTERN_SHAREDMEMORY = c_uint8(4)

WASMTIME_FUNCREF_NULL = (1 << 64) - 1

WASMTIME_COMPONENT_ITEM_COMPONENT = c_uint8(0)
WASMTIME_COMPONENT_ITEM_COMPONENT_INSTANCE = c_uint8(1)
WASMTIME_COMPONENT_ITEM_MODULE = c_uint8(2)
WASMTIME_COMPONENT_ITEM_COMPONENT_FUNC = c_uint8(3)
WASMTIME_COMPONENT_ITEM_RESOURCE = c_uint8(4)
WASMTIME_COMPONENT_ITEM_CORE_FUNC = c_uint8(5)
WASMTIME_COMPONENT_ITEM_TYPE = c_uint8(6)

WASMTIME_COMPONENT_VALTYPE_BOOL = c_uint8(0)
WASMTIME_COMPONENT_VALTYPE_S8 = c_uint8(1)
WASMTIME_COMPONENT_VALTYPE_S16 = c_uint8(2)
WASMTIME_COMPONENT_VALTYPE_S32 = c_uint8(3)
WASMTIME_COMPONENT_VALTYPE_S64 = c_uint8(4)
WASMTIME_COMPONENT_VALTYPE_U8 = c_uint8(5)
WASMTIME_COMPONENT_VALTYPE_U16 = c_uint8(6)
WASMTIME_COMPONENT_VALTYPE_U32 = c_uint8(7)
WASMTIME_COMPONENT_VALTYPE_U64 = c_uint8(8)
WASMTIME_COMPONENT_VALTYPE_F32 = c_uint8(9)
WASMTIME_COMPONENT_VALTYPE_F64 = c_uint8(10)
WASMTIME_COMPONENT_VALTYPE_CHAR = c_uint8(11)
WASMTIME_COMPONENT_VALTYPE_STRING = c_uint8(12)
WASMTIME_COMPONENT_VALTYPE_LIST = c_uint8(13)
WASMTIME_COMPONENT_VALTYPE_RECORD = c_uint8(14)
WASMTIME_COMPONENT_VALTYPE_TUPLE = c_uint8(15)
WASMTIME_COMPONENT_VALTYPE_VARIANT = c_uint8(16)
WASMTIME_COMPONENT_VALTYPE_ENUM = c_uint8(17)
WASMTIME_COMPONENT_VALTYPE_OPTION = c_uint8(18)
WASMTIME_COMPONENT_VALTYPE_RESULT = c_uint8(19)
WASMTIME_COMPONENT_VALTYPE_FLAGS = c_uint8(20)
WASMTIME_COMPONENT_VALTYPE_OWN = c_uint8(21)
WASMTIME_COMPONENT_VALTYPE_BORROW = c_uint8(22)
WASMTIME_COMPONENT_VALTYPE_FUTURE = c_uint8(23)
WASMTIME_COMPONENT_VALTYPE_STREAM = c_uint8(24)
WASMTIME_COMPONENT_VALTYPE_ERROR_CONTEXT = c_uint8(25)

WASMTIME_COMPONENT_BOOL = c_uint8(0)
WASMTIME_COMPONENT_S8 = c_uint8(1)
WASMTIME_COMPONENT_U8 = c_uint8(2)
WASMTIME_COMPONENT_S16 = c_uint8(3)
WASMTIME_COMPONENT_U16 = c_uint8(4)
WASMTIME_COMPONENT_S32 = c_uint8(5)
WASMTIME_COMPONENT_U32 = c_uint8(6)
WASMTIME_COMPONENT_S64 = c_uint8(7)
WASMTIME_COMPONENT_U64 = c_uint8(8)
WASMTIME_COMPONENT_F32 = c_uint8(9)
WASMTIME_COMPONENT_F64 = c_uint8(10)
WASMTIME_COMPONENT_CHAR = c_uint8(11)
WASMTIME_COMPONENT_STRING = c_uint8(12)
WASMTIME_COMPONENT_LIST = c_uint8(13)
WASMTIME_COMPONENT_RECORD = c_uint8(14)
WASMTIME_COMPONENT_TUPLE = c_uint8(15)
WASMTIME_COMPONENT_VARIANT = c_uint8(16)
WASMTIME_COMPONENT_ENUM = c_uint8(17)
WASMTIME_COMPONENT_OPTION = c_uint8(18)
WASMTIME_COMPONENT_RESULT = c_uint8(19)
WASMTIME_COMPONENT_FLAGS = c_uint8(20)
WASMTIME_COMPONENT_RESOURCE = c_uint8(21)

class wasm_ref_t(Structure):
    pass


class wasm_val_union(Union):
    _fields_ = [
        ("i32", c_int32),
        ("i64", c_int64),
        ("f32", c_float),
        ("f64", c_double),
        ("ref", POINTER(wasm_ref_t)),
    ]

    i32: int
    i64: int
    f32: float
    f64: float
    ref: "typing.Union[ctypes._Pointer[wasm_ref_t], None]"


class wasm_val_t(Structure):
    _fields_ = [("kind", c_uint8), ("of", wasm_val_union)]

    kind: int
    of: wasm_val_union


from ._bindings import *  # noqa


def to_bytes(vec: wasm_byte_vec_t) -> bytearray:
    ty = c_uint8 * vec.size
    return bytearray(ty.from_address(addressof(vec.data.contents)))

def to_str(vec: wasm_byte_vec_t) -> str:
    return to_bytes(vec).decode("utf-8")


def to_str_raw(ptr: "ctypes._Pointer", size: int) -> str:
    return string_at(ptr, size).decode("utf-8")


def str_to_capi(s: str) -> wasm_byte_vec_t:
    if not isinstance(s, str):
        raise TypeError("expected a string")
    return bytes_to_capi(s.encode('utf8'))

def bytes_to_capi(s: typing.Union[bytes, bytearray]) -> wasm_byte_vec_t:
    if not isinstance(s, (bytes, bytearray)):
        raise TypeError("expected bytes or bytearray")
    vec = wasm_byte_vec_t()
    wasm_byte_vec_new_uninitialized(byref(vec), len(s))
    buf = (c_uint8 * len(s)).from_buffer_copy(s)
    ctypes.memmove(vec.data, buf, len(s))
    return vec

def take_pointer(structure: ctypes._Pointer, field_name: str) -> ctypes._Pointer:
    """
    Moral equivalent of `mem::replace(&mut structure.field_name, NULL)`

    Ctypes explicitly documents "Surprises" which includes, for example:

            import ctypes

            class A(ctypes.Structure):
                _fields_ = [("x", ctypes.POINTER(ctypes.c_int))]

            x_p = ctypes.pointer(ctypes.c_int(3))
            a = A(x_p)
            x = a.x

            print(x.contents)
            a.x = None
            print(x.contents)

    This program will segfault on the second access. It turns out that `x = a.x`
    is still actually a pointer into the original structure, and `a.x`
    overwrites that field so accessing `x` later accesses null memory. This
    method is an attempt to work around this surprising behavior and actually
    read the field from a structure and replace it with null.

    I'll be honest I just sat through a 3 hour flight, a 5 hour layover, a 9
    hour flight, 1 hour train ride, and I'm sitting in a hotel lobby for
    another 5 hours. That's my state of mind writing this up, so please
    draw conclusions about this method as appropriate.
    """
    field = getattr(structure, field_name)
    assert(isinstance(field, ctypes._Pointer))
    ret = ctypes.cast(ctypes.addressof(field.contents), ctypes.POINTER(field._type_))
    setattr(structure, field_name, None)
    return ret



from .. import _ffi as ffi, WasmtimeError
from ctypes import POINTER
import ctypes
from wasmtime import Managed


class ResourceType(Managed["ctypes._Pointer[ffi.wasmtime_component_resource_type_t]"]):
    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `ResourceType`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_resource_type_t]") -> None:
        ffi.wasmtime_component_resource_type_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_resource_type_t]") -> "ResourceType":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_resource_type_t)):
            raise TypeError("wrong pointer type")
        ty: "ResourceType" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @classmethod
    def host(cls, val: int) -> "ResourceType":
        """
        Creates a new `ResourceType` representing a host-defined resource
        identified by the integer identifier `val` given.
        """
        ptr = ffi.wasmtime_component_resource_type_new_host(val)
        return cls._from_ptr(ptr)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ResourceType):
            return False
        return ffi.wasmtime_component_resource_type_equal(self.ptr(), other.ptr())



import unittest

from wasmtime import Engine, WasmtimeError, Module, Store
from wasmtime.component import Linker, Component, LinkerInstance


class TestLinker(unittest.TestCase):
    def test_smoke(self):
        engine = Engine()
        linker = Linker(engine)
        linker.allow_shadowing = True
        linker.allow_shadowing = False

    def test_init_linker_instance(self):
        with self.assertRaises(WasmtimeError):
            LinkerInstance()

    def test_root(self):
        engine = Engine()
        linker = Linker(engine)
        with linker.root() as root:
            pass
        with linker.root() as root:
            with self.assertRaises(WasmtimeError):
                linker.root()
        with linker.root() as root:
            pass

    def test_add_instance(self):
        engine = Engine()
        linker = Linker(engine)
        with linker.root() as root:
            with root.add_instance('x'):
                pass
            with root.add_instance('y'):
                pass
            with self.assertRaises(WasmtimeError):
                root.add_instance('x')
            with self.assertRaises(WasmtimeError):
                root.add_instance('y')
            with root.add_instance('z'):
                pass

        with linker.root() as root:
            with root.add_instance('again'):
                pass

        with linker.root() as root:
            with root.add_instance('another') as x:
                # Reuse `root` while `x` is alive
                with self.assertRaises(WasmtimeError):
                    root.add_instance('y')

                # Reuse `x` while `x` is alive
                with x.add_instance('x'):
                    with self.assertRaises(WasmtimeError):
                        x.add_instance('y')

    def test_add_module(self):
        engine = Engine()
        linker = Linker(engine)
        with linker.root() as root:
            root.add_module('x', Module(engine, b'(module)'))
            root.add_module('y', Module(engine, b'(module)'))

            with self.assertRaises(WasmtimeError):
                root.add_module('x', Module(engine, b'(module)'))

            with root.add_instance('z'):
                with self.assertRaises(WasmtimeError):
                    root.add_module('not-used-yet', Module(engine, b'(module)'))

    def test_add_wasip2(self):
        engine = Engine()
        linker = Linker(engine)
        linker.add_wasip2()

        with linker.root():
            with self.assertRaises(WasmtimeError):
                linker.add_wasip2()

        linker.close()

        with Linker(engine) as l2:
            l2.add_wasip2()
            with self.assertRaises(WasmtimeError):
                l2.add_wasip2()
            l2.allow_shadowing = True
            l2.add_wasip2()

    def test_host_exception(self):
        engine = Engine()
        store = Store(engine)
        component = Component(engine, """
            (component
                (import "x" (func $x))
                (core module $a
                    (import "" "x" (func))
                    (func (export "x") call 0)
                )
                (core func $x (canon lower (func $x)))
                (core instance $a (instantiate $a
                    (with "" (instance
                        (export "x" (func $x))
                    ))
                ))
                (func (export "x") (canon lift (core func $a "x")))
            )
        """)
        def host_func(_store):
            raise RuntimeError("oh no")
        linker = Linker(engine)
        with linker.root() as l:
            l.add_func('x', host_func)
        instance = linker.instantiate(store, component)
        func = instance.get_func(store, 'x')
        assert(func is not None)
        with self.assertRaises(RuntimeError) as cm:
            func(store)
        self.assertEqual(str(cm.exception), "oh no")

    def test_fail_instantiate(self):
        engine = Engine()
        store = Store(engine)
        component = Component(engine, """
            (component
                (import "x" (func $x))
            )
        """)
        linker = Linker(engine)
        with self.assertRaises(WasmtimeError):
            linker.instantiate(store, component)

    def test_shadow_func(self):
        engine = Engine()
        store = Store(engine)
        linker = Linker(engine)
        with linker.root() as l:
            l.add_func('x', lambda: None)
            with self.assertRaises(WasmtimeError):
                l.add_func('x', lambda: None)
        linker.allow_shadowing = True
        with linker.root() as l:
            l.add_func('x', lambda: None)

    def test_define_unknown_imports_as_traps(self):
        engine = Engine()
        store = Store(engine)
        component = Component(engine, """
            (component
                (import "x" (func $x))
            )
        """)
        linker = Linker(engine)
        with self.assertRaises(WasmtimeError):
            linker.instantiate(store, component)
        linker.define_unknown_imports_as_traps(component)
        linker.instantiate(store, component)



from . import _ffi as ffi
from wasmtime import Config, WasmtimeError, Managed
from typing import Optional
import ctypes


class Engine(Managed["ctypes._Pointer[ffi.wasm_engine_t]"]):

    def __init__(self, config: Optional[Config] = None):
        if config is None:
            self._set_ptr(ffi.wasm_engine_new())
        elif not isinstance(config, Config):
            raise TypeError("expected Config")
        else:
            ptr = config._consume()
            self._set_ptr(ffi.wasm_engine_new_with_config(ptr))

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasm_engine_t]") -> None:
        ffi.wasm_engine_delete(ptr)

    def increment_epoch(self) -> None:
        ffi.wasmtime_engine_increment_epoch(self.ptr())

    def is_pulley(self) -> bool:
        return ffi.wasmtime_engine_is_pulley(self.ptr())



from .. import _ffi as ffi, WasmtimeError, Storelike
from ctypes import byref, pointer
import ctypes
from ._types import ValType, valtype_from_ptr, FuncType
from typing import List, Tuple, Optional, Any
from ._enter import enter_wasm


class Func:
    _func: ffi.wasmtime_component_func_t

    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `Func`")

    @classmethod
    def _from_raw(cls, func: ffi.wasmtime_component_func_t) -> "Func":
        ty: "Func" = cls.__new__(cls)
        ty._func = func
        return ty

    def __call__(self, store: Storelike, *params: Any) -> Any:
        fty = self.type(store)
        param_tys = fty.params
        result_ty = fty.result
        if len(params) != len(param_tys):
            raise TypeError("wrong number of parameters provided: given %s, expected %s" %
                                (len(params), len(param_tys)))
        param_capi = (ffi.wasmtime_component_val_t * len(params))()
        n = 0
        try:
            for (_name, ty), val in zip(param_tys, params):
                ty.convert_to_c(store, val, pointer(param_capi[n]))
                n += 1
            result_space = None
            result_capi = None
            result_len = 0
            if result_ty is not None:
                result_space = ffi.wasmtime_component_val_t()
                result_capi = byref(result_space)
                result_len = 1

            def run() -> 'ctypes._Pointer[ffi.wasmtime_error_t]':
                return ffi.wasmtime_component_func_call(byref(self._func),
                                                        store._context(),
                                                        param_capi,
                                                        n,
                                                        result_capi,
                                                        result_len)
            enter_wasm(run)

            if result_space is None:
                return None
            assert(result_ty is not None)
            return result_ty.convert_from_c(result_space)

        finally:
            for i in range(n):
                ffi.wasmtime_component_val_delete(byref(param_capi[i]))


    def post_return(self, store: Storelike) -> None:
        """
        Performs any necessary post-return operations for this function.
        """
        def run() -> 'ctypes._Pointer[ffi.wasmtime_error_t]':
            return ffi.wasmtime_component_func_post_return(byref(self._func), store._context())
        enter_wasm(run)

    def type(self, store: Storelike) -> FuncType:
        """
        Returns the type of this function.
        """
        ptr = ffi.wasmtime_component_func_type(byref(self._func), store._context())
        return FuncType._from_ptr(ptr)



from . import _ffi as ffi
import ctypes
from wasmtime import WasmtimeError, Managed
import typing


def setter_property(fset: typing.Callable) -> property:
    prop = property(fset=fset)
    if fset.__doc__:
        prop.__doc__ = fset.__doc__
        prop.__doc__ += "\n\n        Note that this field can only be set, it cannot be read"
    return prop


class Config(Managed["ctypes._Pointer[ffi.wasm_config_t]"]):
    """
    Global configuration, used to create an `Engine`.

    A `Config` houses a number of configuration options which tweaks how wasm
    code is compiled or generated.
    """

    def __init__(self) -> None:
        self._set_ptr(ffi.wasm_config_new())

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasm_config_t]") -> None:
        ffi.wasm_config_delete(ptr)

    @setter_property
    def debug_info(self, enable: bool) -> None:
        """
        Configures whether DWARF debug information is emitted for the generated
        code. This can improve profiling and the debugging experience.
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_debug_info_set(self.ptr(), enable)

    @setter_property
    def wasm_threads(self, enable: bool) -> None:
        """
        Configures whether the wasm [threads proposal] is enabled.

        [threads proposal]: https://github.com/webassembly/threads
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_threads_set(self.ptr(), enable)

    @setter_property
    def wasm_tail_call(self, enable: bool) -> None:
        """
        Configures whether the wasm [tail call proposal] is enabled.

        [tail call proposal]: https://github.com/WebAssembly/tail-call
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_tail_call_set(self.ptr(), enable)

    @setter_property
    def wasm_reference_types(self, enable: bool) -> None:
        """
        Configures whether the wasm [reference types proposal] is enabled.

        [reference types proposal]: https://github.com/webassembly/reference-types
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_reference_types_set(self.ptr(), enable)

    @setter_property
    def wasm_simd(self, enable: bool) -> None:
        """
        Configures whether the wasm [SIMD proposal] is enabled.

        [SIMD proposal]: https://github.com/webassembly/simd
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_simd_set(self.ptr(), enable)

    @setter_property
    def wasm_bulk_memory(self, enable: bool) -> None:
        """
        Configures whether the wasm [bulk memory proposal] is enabled.

        [bulk memory proposal]: https://github.com/webassembly/bulk-memory
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_bulk_memory_set(self.ptr(), enable)

    @setter_property
    def wasm_multi_value(self, enable: bool) -> None:
        """
        Configures whether the wasm [multi value proposal] is enabled.

        [multi value proposal]: https://github.com/webassembly/multi-value
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_multi_value_set(self.ptr(), enable)

    @setter_property
    def wasm_multi_memory(self, enable: bool) -> None:
        """
        Configures whether the wasm [multi memory proposal] is enabled.

        [multi memory proposal]: https://github.com/webassembly/multi-memory
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_multi_memory_set(self.ptr(), enable)

    @setter_property
    def wasm_memory64(self, enable: bool) -> None:
        """
        Configures whether the wasm [memory64 proposal] is enabled.

        [memory64 proposal]: https://github.com/webassembly/memory64
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_memory64_set(self.ptr(), enable)

    @setter_property
    def wasm_relaxed_simd(self, enable: bool) -> None:
        """
        Configures whether the wasm [relaxed simd proposal] is enabled.

        [relaxed simd proposal]: https://github.com/webassembly/relaxed-simd
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_relaxed_simd_set(self.ptr(), enable)

    @setter_property
    def wasm_relaxed_simd_deterministic(self, enable: bool) -> None:
        """
        Configures whether the wasm [relaxed simd proposal] is deterministic
        in is execution as opposed to having the most optimal implementation for
        the current platform.

        [relaxed simd proposal]: https://github.com/webassembly/relaxed-simd
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_relaxed_simd_deterministic_set(self.ptr(), enable)

    @setter_property
    def wasm_exceptions(self, enable: bool) -> None:
        """
        Configures whether the wasm [exceptions proposal] is enabled.

        [exceptions proposal]: https://github.com/WebAssembly/exception-handling
        """

        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_wasm_exceptions_set(self.ptr(), enable)

    @setter_property
    def strategy(self, strategy: str) -> None:
        """
        Configures the compilation strategy used for wasm code.

        Acceptable values for `strategy` are:

        * `"auto"`
        * `"cranelift"`
        """

        if strategy == "auto":
            ffi.wasmtime_config_strategy_set(self.ptr(), 0)
        elif strategy == "cranelift":
            ffi.wasmtime_config_strategy_set(self.ptr(), 1)
        else:
            raise WasmtimeError("unknown strategy: " + str(strategy))

    @setter_property
    def cranelift_debug_verifier(self, enable: bool) -> None:
        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_cranelift_debug_verifier_set(self.ptr(), enable)

    @setter_property
    def cranelift_opt_level(self, opt_level: str) -> None:
        if opt_level == "none":
            ffi.wasmtime_config_cranelift_opt_level_set(self.ptr(), 0)
        elif opt_level == "speed":
            ffi.wasmtime_config_cranelift_opt_level_set(self.ptr(), 1)
        elif opt_level == "speed_and_size":
            ffi.wasmtime_config_cranelift_opt_level_set(self.ptr(), 2)
        else:
            raise WasmtimeError("unknown opt level: " + str(opt_level))

    @setter_property
    def profiler(self, profiler: str) -> None:
        if profiler == "none":
            ffi.wasmtime_config_profiler_set(self.ptr(), 0)
        elif profiler == "jitdump":
            ffi.wasmtime_config_profiler_set(self.ptr(), 1)
        else:
            raise WasmtimeError("unknown profiler: " + str(profiler))

    @setter_property
    def cache(self, enabled: typing.Union[bool, str]) -> None:
        """
        Configures whether code caching is enabled for this `Config`.

        The value `True` can be passed in here to enable the default caching
        configuration and location, or a path to a file can be passed in which
        is a path to a TOML configuration file for the cache.

        More information about cache configuration can be found at
        https://bytecodealliance.github.io/wasmtime/cli-cache.html
        """

        if isinstance(enabled, bool):
            if not enabled:
                raise WasmtimeError("caching cannot be explicitly disabled")
            error = ffi.wasmtime_config_cache_config_load(self.ptr(), None)
        elif isinstance(enabled, str):
            error = ffi.wasmtime_config_cache_config_load(self.ptr(),
                ctypes.c_char_p(enabled.encode('utf-8')),
            )
        else:
            raise TypeError("expected string or bool")
        if error:
            raise WasmtimeError._from_ptr(error)

    @setter_property
    def epoch_interruption(self, enabled: bool) -> None:
        """
        Configures whether wasm execution can be interrupted via epoch
        increments.
        """

        if enabled:
            val = 1
        else:
            val = 0
        ffi.wasmtime_config_epoch_interruption_set(self.ptr(), val)

    @setter_property
    def consume_fuel(self, instances: bool) -> None:
        """
        Configures whether wasm code will consume *fuel* as part of its
        execution.

        Fuel consumption allows WebAssembly to trap when fuel runs out.
        Currently stores start with 0 fuel if this is enabled.
        """
        if not isinstance(instances, bool):
            raise TypeError('expected an bool')
        ffi.wasmtime_config_consume_fuel_set(self.ptr(), instances)

    @setter_property
    def parallel_compilation(self, enable: bool) -> None:
        """
        Configures whether parallel compilation is enabled for functions
        within a module.

        This is enabled by default.
        """
        if not isinstance(enable, bool):
            raise TypeError('expected a bool')
        ffi.wasmtime_config_parallel_compilation_set(self.ptr(), enable)



# This is an example of working with multi-value modules and dealing with
# multi-value functions.

from wasmtime import Config, Store, Engine, Module, FuncType, Func, ValType, Instance

def run():
    # Configure our `Store`, but be sure to use a `Config` that enables the
    # wasm multi-value feature since it's not stable yet.
    print("Initializing...")
    config = Config()
    config.wasm_multi_value = True
    store = Store(Engine(config))

    print("Compiling module...")
    module = Module.from_file(store.engine, "examples/multi.wat")

    print("Creating callback...")
    callback_type = FuncType([ValType.i32(), ValType.i64()], [ValType.i64(), ValType.i32()])


    def callback(a, b):
        return [b + 1, a + 1]


    callback_func = Func(store, callback_type, callback)

    print("Instantiating module...")
    instance = Instance(store, module, [callback_func])

    print("Extracting export...")
    g = instance.exports(store)["g"]

    print("Calling export \"g\"...")
    results = g(store, 1, 3)
    print("> {} {}".format(results[0], results[1]))

    assert(results[0] == 4)
    assert(results[1] == 2)

    print("Calling export \"round_trip_many\"...")
    round_trip_many = instance.exports(store)["round_trip_many"]
    results = round_trip_many(store, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9)

    print("Printing result...")
    print(">")
    for r in results:
        print("  %d" % r)
    assert(len(results) == 10)
    for i, r in enumerate(results):
        assert(i == r)

if __name__ == '__main__':
    run()



import unittest
import tempfile

from wasmtime import *
from wasmtime.component import Component, ExportIndex

class TestComponent(unittest.TestCase):
    def test_smoke(self):
        Component(Engine(), '(component)')
        Component(Engine(), bytes(b'\0asm\x0d\0\x01\0'))
        Component(Engine(), bytearray(b'\0asm\x0d\0\x01\0'))

        with self.assertRaises(WasmtimeError):
            Component(Engine(), '(component2)')
        with self.assertRaises(WasmtimeError):
            Component(Engine(), bytes(b'\0asm\x01\0\0\0'))

    def test_invalid(self):
        with self.assertRaises(TypeError):
            Component(1, b'')  # type: ignore
        with self.assertRaises(TypeError):
            Component(Engine(), 2)  # type: ignore
        with self.assertRaises(WasmtimeError):
            Component(Engine(), b'')
        with self.assertRaises(WasmtimeError):
            Component(Engine(), b'\x00')

    def test_serialize(self):
        engine = Engine()
        component = Component(engine, '(component)')
        encoded = component.serialize()
        component = Component.deserialize(engine, encoded)
        with tempfile.TemporaryDirectory() as d:
            path = d + '/component.bin'
            with open(path, 'wb') as f:
                f.write(encoded)
            # Run the destructor for `Component` which has an mmap to the file
            # which prevents deletion on Windows.
            with Component.deserialize_file(engine, path):
                pass

    def test_exports(self):
        engine = Engine()

        c = Component(engine, '(component)')
        self.assertIsNone(c.get_export_index('foo'))
        self.assertIsNone(c.get_export_index('foo', instance = None))

        c = Component(engine, """
            (component
                (core module (export "foo"))
            )
        """)
        foo = c.get_export_index('foo')
        self.assertIsNotNone(foo)
        self.assertIsNone(c.get_export_index('foo', instance = foo))
        self.assertIsInstance(foo, ExportIndex)

        c = Component(engine, """
            (component
                (core module $a)
                (instance (export "x")
                    (export "m" (core module $a))
                )
            )
        """)
        self.assertIsNotNone(c.get_export_index('x'))
        self.assertIsNotNone(c.get_export_index('m', instance = c.get_export_index('x')))

        c2 = Component(engine, """
            (component
                (core module $a)
                (instance (export "x")
                    (export "m" (core module $a))
                )
            )
        """)
        self.assertIsNone(c2.get_export_index('m', instance = c.get_export_index('x')))



import unittest

from wasmtime import *


class TestEngine(unittest.TestCase):
    def test_smoke(self):
        Engine()
        Engine(Config())

        Engine().is_pulley()

    def test_errors(self):
        with self.assertRaises(TypeError):
            Engine(3)  # type: ignore
        config = Config()
        Engine(config)
        with self.assertRaises(ValueError):
            Engine(config)



import unittest

from wasmtime import Engine, FuncType as CoreFuncType, GlobalType, WasmtimeError
from wasmtime.component import *

def simplety(name):
    engine = Engine()
    c = Component(engine, f"""
    (component
      (import "a" (func (result {name})))
    )
    """)
    fty = c.type.imports(engine)['a']
    assert(isinstance(fty, FuncType))
    return fty.result

def namedty(contents):
    engine = Engine()
    c = Component(engine, f"""
    (component
      (type $t' {contents})
      (import "t" (type $t (eq $t')))
      (import "a" (func (result $t)))
    )
    """)
    fty = c.type.imports(engine)['a']
    assert(isinstance(fty, FuncType))
    return fty.result

class TestTypes(unittest.TestCase):
    def test_component(self):
        engine = Engine()
        ty = Component(engine, '(component)').type
        self.assertEqual(len(ty.imports(engine)), 0)
        self.assertEqual(len(ty.exports(engine)), 0)

        ty = Component(engine, """
        (component
          (import "a" (core module $a))
          (export "g" (core module $a))
        )
        """).type
        self.assertEqual(len(ty.imports(engine)), 1)
        self.assertIsInstance(ty.imports(engine)['a'], ModuleType)
        self.assertEqual(len(ty.exports(engine)), 1)
        self.assertIsInstance(ty.exports(engine)['g'], ModuleType)

        with self.assertRaises(WasmtimeError):
            ComponentType()

    def test_module(self):
        engine = Engine()
        c = Component(engine, """
        (component
          (import "a" (core module $a
            (import "b" "c" (func))
            (export "d" (global i32))
          ))
        )
        """)
        mty = c.type.imports(engine)['a']
        assert(isinstance(mty, ModuleType))
        imports = mty.imports(engine)
        self.assertEqual(len(imports), 1)
        exports = mty.exports(engine)
        self.assertEqual(len(exports), 1)

        self.assertEqual(imports[0].module, 'b')
        self.assertEqual(imports[0].name, 'c')
        self.assertIsInstance(imports[0].type, CoreFuncType)

        self.assertEqual(exports[0].name, 'd')
        self.assertIsInstance(exports[0].type, GlobalType)

        with self.assertRaises(WasmtimeError):
            ModuleType()

    def test_resource(self):
        engine = Engine()
        c = Component(engine, """
        (component
          (import "a" (type $t (sub resource)))
          (export "a" (type $t))
          (type $t2 (resource (rep i32)))
          (export "b" (type $t2))
        )
        """)
        a1 = c.type.imports(engine)['a']
        a2 = c.type.exports(engine)['a']
        b = c.type.exports(engine)['b']
        assert(isinstance(a1, ResourceType))
        assert(isinstance(a2, ResourceType))
        assert(isinstance(b, ResourceType))
        self.assertEqual(a1, a2)
        self.assertNotEqual(a1, b)
        self.assertNotEqual(a2, b)
        self.assertNotEqual(a2, 'hello')

        with self.assertRaises(WasmtimeError):
            ResourceType()

    def test_instance(self):
        engine = Engine()
        cty = Component(engine, """
        (component
          (import "a" (instance
            (export "a" (func))
            (export "b" (core module))
          ))
          (import "b" (instance))
        )
        """).type
        a = cty.imports(engine)['a']
        b = cty.imports(engine)['b']
        assert(isinstance(a, ComponentInstanceType))
        assert(isinstance(b, ComponentInstanceType))
        exports = a.exports(engine)
        self.assertEqual(len(exports), 2)
        self.assertIsInstance(exports['a'], FuncType)
        self.assertIsInstance(exports['b'], ModuleType)

        self.assertEqual(len(b.exports(engine)), 0)

        with self.assertRaises(WasmtimeError):
            ComponentInstanceType()

    def test_func(self):
        engine = Engine()
        cty = Component(engine, """
        (component
          (import "a" (func))
          (import "b" (func (param "x" u32)))
          (import "c" (func (result string)))
          (import "d" (func (param "a" u8) (param "b" u16) (result u32)))
        )
        """).type
        a = cty.imports(engine)['a']
        b = cty.imports(engine)['b']
        c = cty.imports(engine)['c']
        d = cty.imports(engine)['d']
        assert(isinstance(a, FuncType))
        assert(isinstance(b, FuncType))
        assert(isinstance(c, FuncType))
        assert(isinstance(d, FuncType))

        self.assertEqual(a.params, [])
        self.assertIsNone(a.result)
        self.assertEqual(b.params, [('x', U32())])
        self.assertIsNone(a.result)
        self.assertEqual(c.params, [])
        self.assertEqual(c.result, String())
        self.assertEqual(d.params, [('a', U8()), ('b', U16())])
        self.assertEqual(d.result, U32())

        with self.assertRaises(WasmtimeError):
            FuncType()

    def test_primitives(self):
        self.assertEqual(simplety('bool'), Bool())
        self.assertEqual(simplety('u8'), U8())
        self.assertEqual(simplety('u16'), U16())
        self.assertEqual(simplety('u32'), U32())
        self.assertEqual(simplety('u64'), U64())
        self.assertEqual(simplety('s8'), S8())
        self.assertEqual(simplety('s16'), S16())
        self.assertEqual(simplety('s32'), S32())
        self.assertEqual(simplety('s64'), S64())
        self.assertEqual(simplety('f32'), F32())
        self.assertEqual(simplety('f64'), F64())
        self.assertEqual(simplety('char'), Char())
        self.assertEqual(simplety('string'), String())

    def test_list(self):
        l = simplety('(list u8)')
        assert(isinstance(l, ListType))
        self.assertEqual(l.element, U8())
        self.assertEqual(l, simplety('(list u8)'))
        self.assertNotEqual(l, simplety('(list u16)'))

    def test_record(self):
        r = namedty('(record (field "a" u8) (field "b" f32))')
        assert(isinstance(r, RecordType))
        self.assertEqual(r.fields, [('a', U8()), ('b', F32())])
        self.assertEqual(r, namedty('(record (field "a" u8) (field "b" f32))'))
        self.assertNotEqual(r, namedty('(record (field "a" u8) (field "b" f64))'))

    def test_tuple(self):
        t = namedty('(tuple u8 f32)')
        assert(isinstance(t, TupleType))
        self.assertEqual(t.elements, [U8(), F32()])
        self.assertEqual(t, namedty('(tuple u8 f32)'))
        self.assertNotEqual(t, namedty('(tuple u8 f64)'))

    def test_variant(self):
        t = namedty('(variant (case "a") (case "b" f32))')
        assert(isinstance(t, VariantType))
        self.assertEqual(t.cases, [('a', None), ('b', F32())])
        self.assertEqual(t, namedty('(variant (case "a") (case "b" f32))'))
        self.assertNotEqual(t, namedty('(variant (case "a") (case "b" f64))'))

    def test_enum(self):
        e = namedty('(enum "a" "b" "c")')
        assert(isinstance(e, EnumType))
        self.assertEqual(e.names, ['a', 'b', 'c'])
        self.assertEqual(e, namedty('(enum "a" "b" "c")'))
        self.assertNotEqual(e, namedty('(enum "a" "b" "d")'))

    def test_option(self):
        o = simplety('(option u32)')
        assert(isinstance(o, OptionType))
        self.assertEqual(o.payload, U32())
        self.assertEqual(o, simplety('(option u32)'))
        self.assertNotEqual(o, simplety('(option u64)'))

    def test_result(self):
        r = simplety('(result u32 (error f32))')
        assert(isinstance(r, ResultType))
        self.assertEqual(r.ok, U32())
        self.assertEqual(r.err, F32())
        self.assertEqual(r, simplety('(result u32 (error f32))'))

        r = simplety('(result (error f32))')
        assert(isinstance(r, ResultType))
        self.assertIsNone(r.ok)
        self.assertEqual(r.err, F32())

        r = simplety('(result u32)')
        assert(isinstance(r, ResultType))
        self.assertEqual(r.ok, U32())
        self.assertIsNone(r.err)

        r = simplety('(result)')
        assert(isinstance(r, ResultType))
        self.assertIsNone(r.ok)
        self.assertIsNone(r.err)

    def test_flags(self):
        f = namedty('(flags "a" "b" "c")')
        assert(isinstance(f, FlagsType))
        self.assertEqual(f.names, ['a', 'b', 'c'])
        self.assertEqual(f, namedty('(flags "a" "b" "c")'))
        self.assertNotEqual(f, namedty('(flags "a" "b" "d")'))

    def test_own_and_borrow(self):
        engine = Engine()
        c = Component(engine, f"""
        (component
            (import "r" (type $r (sub resource)))
            (import "a" (func (param "x" (borrow $r)) (result (own $r))))
        )
        """)
        fty = c.type.imports(engine)['a']
        assert(isinstance(fty, FuncType))
        _, param = fty.params[0]
        result = fty.result
        assert(isinstance(param, BorrowType))
        assert(isinstance(result, OwnType))
        self.assertEqual(param.ty, result.ty)



import ctypes
from typing import Any
from wasmtime import Instance, Engine, FuncType
from wasmtime import Module, WasmtimeError, Func, Managed
from . import _ffi as ffi
from ._extern import get_extern_ptr, wrap_extern
from ._config import setter_property
from ._exportable import AsExtern
from ._store import Storelike
from ._func import enter_wasm, trampoline, FUNCTIONS, finalize
from typing import Callable


class Linker(Managed["ctypes._Pointer[ffi.wasmtime_linker_t]"]):
    engine: Engine

    def __init__(self, engine: Engine):
        """
        Creates a new linker ready to instantiate modules within the store
        provided.
        """
        self._set_ptr(ffi.wasmtime_linker_new(engine.ptr()))
        self.engine = engine

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_linker_t]") -> None:
        ffi.wasmtime_linker_delete(ptr)

    @setter_property
    def allow_shadowing(self, allow: bool) -> None:
        """
        Configures whether definitions are allowed to shadow one another within
        this linker
        """
        ffi.wasmtime_linker_allow_shadowing(self.ptr(), allow)

    def define(self, store: Storelike, module: str, name: str, item: AsExtern) -> None:
        """
        Defines a new item, by name, in this linker.

        This method will add a new definition to this linker. The `module` nad
        `name` provided are what to name the `item` within the linker.

        This function will raise an error if `item` comes from the wrong store
        or if shadowing is disallowed and the module/name pair has already been
        defined.
        """
        raw_item = get_extern_ptr(item)
        module_bytes = module.encode('utf-8')
        module_buf = ctypes.create_string_buffer(module_bytes)
        name_bytes = name.encode('utf-8')
        name_buf = ctypes.create_string_buffer(name_bytes)
        error = ffi.wasmtime_linker_define(
            self.ptr(),
            store._context(),
            module_buf,
            len(module_bytes),
            name_buf,
            len(name_bytes),
            ctypes.byref(raw_item))
        if error:
            raise WasmtimeError._from_ptr(error)

    def define_func(self, module: str, name: str, ty: FuncType, func: Callable[..., Any], access_caller: bool = False) -> None:
        """
        Defines a new function, by name, in this linker.

        This method is similar to `define` except that you can directly define a
        function without creating a `Func` itself. This enables
        `Store`-independent functions to be inserted into this linker, meaning
        the linker can be used to instantiate modules in multiple stores.
        """
        module_bytes = module.encode('utf-8')
        module_buf = ctypes.create_string_buffer(module_bytes)
        name_bytes = name.encode('utf-8')
        name_buf = ctypes.create_string_buffer(name_bytes)
        if not isinstance(ty, FuncType):
            raise TypeError("expected a FuncType")
        idx = FUNCTIONS.allocate((func, ty.results, access_caller))
        error = ffi.wasmtime_linker_define_func(
            self.ptr(),
            module_buf,
            len(module_bytes),
            name_buf,
            len(name_bytes),
            ty.ptr(),
            trampoline,
            idx,
            finalize)
        if error:
            raise WasmtimeError._from_ptr(error)

    def define_instance(self, store: Storelike, name: str, instance: Instance) -> None:
        """
        Convenience wrapper to define an entire instance in this linker.

        This function will `define` eaech of the exports on the instance into
        this linker, using the name provided as the module name and the export's
        own name as the field name.

        This function will raise an error if `instance` comes from the wrong
        store or if shadowing is disallowed and a name was previously defined.
        """
        if not isinstance(instance, Instance):
            raise TypeError("expected an `Instance`")
        name_bytes = name.encode('utf8')
        name_buf = ctypes.create_string_buffer(name_bytes)
        error = ffi.wasmtime_linker_define_instance(self.ptr(),
                                                    store._context(),
                                                    name_buf,
                                                    len(name_bytes),
                                                    ctypes.byref(instance._instance))
        if error:
            raise WasmtimeError._from_ptr(error)

    def define_wasi(self) -> None:
        """
        Defines a WASI instance in this linker.

        The instance provided has been previously constructed and this method
        will define all the appropriate imports and their names into this linker
        to assist with instantiating modules that use WASI.

        This function will raise an error if shadowing is disallowed and a name
        was previously defined.
        """
        error = ffi.wasmtime_linker_define_wasi(self.ptr())
        if error:
            raise WasmtimeError._from_ptr(error)

    def define_module(self, store: Storelike, name: str, module: Module) -> None:
        """
        Defines automatic instantiations of the provided module in this linker.

        The `module` provided is defined under `name` with automatic
        instantiations which respect WASI Commands and Reactors.

        For more information see the Rust documentation at
        https://docs.wasmtime.dev/api/wasmtime/struct.Linker.html#method.module.

        This method will throw an error if shadowing is disallowed and an item
        has previously been defined.
        """
        if not isinstance(module, Module):
            raise TypeError("expected a `Module`")
        name_bytes = name.encode('utf-8')
        name_buf = ctypes.create_string_buffer(name_bytes)
        error = ffi.wasmtime_linker_module(self.ptr(), store._context(), name_buf, len(name_bytes), module.ptr())
        if error:
            raise WasmtimeError._from_ptr(error)

    def instantiate(self, store: Storelike, module: Module) -> Instance:
        """
        Instantiates a module using this linker's defined set of names.

        This method will attempt to satisfy all the imports of the `module`
        provided with the names defined within this linker. If all names are
        defined then the module is instantiated.

        Raises an error if an import of `module` hasn't been defined in this
        linker or if a trap happens while instantiating the instance.
        """
        trap = ctypes.POINTER(ffi.wasm_trap_t)()
        instance = ffi.wasmtime_instance_t()
        with enter_wasm(store) as trap:
            error = ffi.wasmtime_linker_instantiate(
                self.ptr(), store._context(), module.ptr(), ctypes.byref(instance), trap)
            if error:
                raise WasmtimeError._from_ptr(error)
        return Instance._from_raw(instance)

    def get_default(self, store: Storelike, name: str) -> Func:
        """
        Gets the default export for the named module in this linker.

        For more information on this see the Rust documentation at
        https://docs.wasmtime.dev/api/wasmtime/struct.Linker.html#method.get_default.

        Raises an error if the default export wasn't present.
        """
        name_bytes = name.encode('utf-8')
        name_buf = ctypes.create_string_buffer(name_bytes)
        func = ffi.wasmtime_func_t()
        error = ffi.wasmtime_linker_get_default(self.ptr(), store._context(),
                                                name_buf, len(name_bytes), ctypes.byref(func))
        if error:
            raise WasmtimeError._from_ptr(error)
        return Func._from_raw(func)

    def get(self, store: Storelike, module: str, name: str) -> AsExtern:
        """
        Gets a singular item defined in this linker.

        Raises an error if this item hasn't been defined or if the item has been
        defined twice with different types.
        """
        module_bytes = module.encode('utf-8')
        module_buf = ctypes.create_string_buffer(module_bytes)
        name_bytes = name.encode('utf-8')
        name_buf = ctypes.create_string_buffer(name_bytes)
        item = ffi.wasmtime_extern_t()
        ok = ffi.wasmtime_linker_get(self.ptr(), store._context(),
                                     module_buf, len(module_bytes),
                                     name_buf, len(name_bytes),
                                     ctypes.byref(item))
        if ok:
            return wrap_extern(item)
        raise WasmtimeError("item not defined in linker")



import ctypes

from .. import _ffi as ffi, WasmtimeError, Storelike
from ctypes import byref
from typing import Optional, Union
from ._component import ExportIndex
from ._func import Func


class Instance:
    _instance: ffi.wasmtime_component_instance_t

    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct an `Instance`")

    @classmethod
    def _from_raw(cls, instance: ffi.wasmtime_component_instance_t) -> "Instance":
        ty: "Instance" = cls.__new__(cls)
        ty._instance = instance
        return ty

    def get_export_index(self, store: Storelike, name: str, instance: Optional[ExportIndex] = None) -> Optional[ExportIndex]:
        """
        Retrieves the export index for the export named `name` from this
        component instance.

        If `instance` is provided then the export is looked up within that
        nested instance.

        Returns `None` if no such export exists otherwise returns the export
        index.
        """
        name_bytes = name.encode('utf-8')
        name_buf = ctypes.create_string_buffer(name_bytes)

        ptr = ffi.wasmtime_component_instance_get_export_index(
            byref(self._instance),
            store._context(),
            instance.ptr() if instance is not None else None,
            name_buf,
            len(name_bytes))
        if ptr:
            return ExportIndex._from_ptr(ptr)
        return None

    def get_func(self, store: Storelike, index: Union[ExportIndex, str]) -> Optional[Func]:
        """
        Retrieves the function export for the given export index.

        Returns `None` if the export index does not correspond to a function
        export and otherwise returns the function.
        """
        if isinstance(index, str):
            ei = self.get_export_index(store, index)
            if ei is None:
                return None
            index = ei

        raw = ffi.wasmtime_component_func_t()
        found = ffi.wasmtime_component_instance_get_func(
            byref(self._instance),
            store._context(),
            index.ptr(),
            byref(raw))
        if found:
            return Func._from_raw(raw)
        return None



import unittest

from wasmtime import *


class TestGlobal(unittest.TestCase):
    def test_new(self):
        store = Store()
        ty = GlobalType(ValType.i32(), True)
        g = Global(store, ty, Val.i32(1))
        Global(store, ty, Val.i32(1))
        self.assertEqual(g.type(store).content, ValType.i32())
        self.assertTrue(g.type(store).mutable)

        self.assertEqual(g.value(store), 1)
        g.set_value(store, Val.i32(2))
        self.assertEqual(g.value(store), 2)
        self.assertTrue(isinstance(g.type(store), GlobalType))

    def test_falsy(self):
        store = Store()
        ty = GlobalType(ValType.i64(), True)
        g = Global(store, ty, Val.i64(0))
        val = g.value(store)

        self.assertEqual(val, 0)
        self.assertTrue(isinstance(val, int))

    def test_errors(self):
        store = Store()
        ty = GlobalType(ValType.i32(), True)
        with self.assertRaises(TypeError):
            Global(store, ty, store)
        with self.assertRaises(TypeError):
            Global(store, 1, Val.i32(1))  # type: ignore
        with self.assertRaises(AttributeError):
            Global(1, ty, Val.i32(1))  # type: ignore

        g = Global(store, ty, Val.i32(1))
        with self.assertRaises(TypeError):
            g.set_value(store, g)

        ty = GlobalType(ValType.i32(), False)
        g = Global(store, ty, Val.i32(1))
        with self.assertRaises(WasmtimeError):
            g.set_value(store, 1)



# Example of instantiating a wasm module and calling an export on it

from wasmtime import Store, Module, Instance

def run():
    store = Store()
    module = Module.from_file(store.engine, './examples/gcd.wat')
    instance = Instance(store, module, [])
    gcd = instance.exports(store)["gcd"]

    print("gcd(6, 27) = %d" % gcd(store, 6, 27))

if __name__ == '__main__':
    run()



from .. import _ffi as ffi, WasmtimeError, Managed, Storelike, WasmtimeError
from ctypes import POINTER, byref
import ctypes
from ._resource_type import ResourceType
from ._enter import enter_wasm


class ResourceAny(Managed["ctypes._Pointer[ffi.wasmtime_component_resource_any_t]"]):
    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `ResourceAny`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_resource_any_t]") -> None:
        ffi.wasmtime_component_resource_any_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_resource_any_t]") -> "ResourceAny":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_resource_any_t)):
            raise TypeError("wrong pointer type")
        ty: "ResourceAny" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @property
    def type(self) -> ResourceType:
        """
        Returns the `ResourceType` of this `ResourceAny`.
        """
        ptr = ffi.wasmtime_component_resource_any_type(self.ptr())
        return ResourceType._from_ptr(ptr)

    @property
    def owned(self) -> bool:
        """
        Returns whether this `ResourceAny` is an `own` resource or a `borrow`
        resource.
        """
        return ffi.wasmtime_component_resource_any_owned(self.ptr())

    def drop(self, store: Storelike) -> None:
        """
        Runs the WebAssembly-defined destructor, if any, for this resource.

        This is required to be called to clean up state information in the
        store about this resource. This may execute WebAssembly code if the
        resource is a guest-owned resource with a defined destructor.
        """
        def run() -> 'ctypes._Pointer[ffi.wasmtime_error_t]':
            return ffi.wasmtime_component_resource_any_drop(store._context(), self.ptr())
        enter_wasm(run)

    def to_host(self, store: Storelike) -> "ResourceHost":
        """
        Attempts to downcast this `ResourceAny` to a `ResourceHost`.

        This will raise a `WasmtimeError` if this `ResourceAny` is not
        actually a host-defined resource.
        """
        ptr = POINTER(ffi.wasmtime_component_resource_host_t)()
        error = ffi.wasmtime_component_resource_any_to_host(store._context(), self.ptr(), byref(ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        return ResourceHost._from_ptr(ptr)


class ResourceHost(Managed["ctypes._Pointer[ffi.wasmtime_component_resource_host_t]"]):
    def __init__(self) -> None:
        raise WasmtimeError("Cannot directly construct a `ResourceHost`")

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_component_resource_host_t]") -> None:
        ffi.wasmtime_component_resource_host_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_component_resource_host_t]") -> "ResourceHost":
        if not isinstance(ptr, POINTER(ffi.wasmtime_component_resource_host_t)):
            raise TypeError("wrong pointer type")
        ty: "ResourceHost" = cls.__new__(cls)
        ty._set_ptr(ptr)
        return ty

    @classmethod
    def own(cls, rep: int, ty: int) -> "ResourceHost":
        ptr = ffi.wasmtime_component_resource_host_new(True, rep, ty)
        return cls._from_ptr(ptr)

    @classmethod
    def borrow(cls, rep: int, ty: int) -> "ResourceHost":
        ptr = ffi.wasmtime_component_resource_host_new(False, rep, ty)
        return cls._from_ptr(ptr)

    @property
    def rep(self) -> int:
        """
        Returns the integer representation associated with this host resource.
        """
        return ffi.wasmtime_component_resource_host_rep(self.ptr())

    @property
    def type(self) -> int:
        """
        Returns the integer type identifier associated with this host resource.
        """
        return ffi.wasmtime_component_resource_host_type(self.ptr())

    @property
    def owned(self) -> bool:
        """
        Returns whether this `ResourceHost` is an `own` resource or a `borrow`
        resource.
        """
        return ffi.wasmtime_component_resource_host_owned(self.ptr())

    def to_any(self, store: Storelike) -> ResourceAny:
        """
        Upcasts this `ResourceHost` to a `ResourceAny`.
        """
        ptr = POINTER(ffi.wasmtime_component_resource_any_t)()
        error = ffi.wasmtime_component_resource_host_to_any(store._context(), self.ptr(), byref(ptr))
        if error:
            raise WasmtimeError._from_ptr(error)
        return ResourceAny._from_ptr(ptr)



from ctypes import byref, POINTER, c_int
from . import _ffi as ffi
import ctypes
from typing import Optional
from wasmtime import Managed


class WasmtimeError(Exception, Managed["ctypes._Pointer[ffi.wasmtime_error_t]"]):
    __message: Optional[str]

    def __init__(self, message: str):
        self.__message = message
        self._set_ptr(ffi.wasmtime_error_new(message.encode('utf-8')))

    def _delete(self, ptr: "ctypes._Pointer[ffi.wasmtime_error_t]") -> None:
        ffi.wasmtime_error_delete(ptr)

    @classmethod
    def _from_ptr(cls, ptr: "ctypes._Pointer[ffi.wasmtime_error_t]") -> 'WasmtimeError':
        from . import _ffi as ffi
        if not isinstance(ptr, POINTER(ffi.wasmtime_error_t)):
            raise TypeError("wrong pointer type")

        exit_code = c_int(0)
        if ffi.wasmtime_error_exit_status(ptr, byref(exit_code)):
            exit_trap: ExitTrap = ExitTrap.__new__(ExitTrap)
            exit_trap._set_ptr(ptr)
            exit_trap.__message = None
            exit_trap.code = exit_code.value
            return exit_trap

        err: WasmtimeError = cls.__new__(cls)
        err._set_ptr(ptr)
        err.__message = None
        return err

    def __str__(self) -> str:
        if self.__message:
            return self.__message
        message_vec = ffi.wasm_byte_vec_t()
        ffi.wasmtime_error_message(self.ptr(), byref(message_vec))
        message = ffi.to_str(message_vec)
        ffi.wasm_byte_vec_delete(byref(message_vec))
        return message


class ExitTrap(WasmtimeError):
    """
    A special type of `WasmtimeError` which represents the process exiting via
    WASI's `proc_exit` function call.

    Whenever a WASI program exits via `proc_exit` a trap is raised, but the
    trap will have this type instead of `WasmtimeError`, so you can catch just
    this type instead of all traps (if desired). Exit traps have a `code`
    associated with them which is the exit code provided at exit.

    Note that `ExitTrap` is a subclass of `WasmtimeError`, so if you catch a
    trap you'll also catch `ExitTrap`.
    """
    code: int
    pass



import unittest
import tempfile

from wasmtime import *
from pathlib import Path


class TestWasi(unittest.TestCase):
    def test_config(self):
        config = WasiConfig()
        config.argv = ['a', 'b']
        config.inherit_argv()
        config.env = [['a', 'b']]
        config.inherit_env()

        with tempfile.NamedTemporaryFile() as f:
            config.stdin_file = f.name
            config.stdin_file = Path(f.name)
            config.inherit_stdin()
            config.stdout_file = f.name
            config.stdout_file = Path(f.name)
            config.inherit_stdout()
            config.stderr_file = f.name
            config.stderr_file = Path(f.name)
            config.inherit_stderr()

        with self.assertRaises(WasmtimeError):
            config.stdin_file = 'somewhere-over-the-rainboat'
        with self.assertRaises(WasmtimeError):
            config.stdout_file = 'some-directory/without-a-rainbow'
        with self.assertRaises(WasmtimeError):
            config.stderr_file = 'some-directory/without-a-rainbow'
        config.preopen_dir('wasmtime', 'other', DirPerms.READ_WRITE, FilePerms.READ_WRITE)
        config.preopen_dir('wasmtime', 'other2')

    def test_preview1(self):
        linker = Linker(Engine())
        linker.define_wasi()

        module = Module(linker.engine, """
            (module
                (import "wasi_snapshot_preview1" "random_get"
                    (func (param i32 i32) (result i32)))
            )
        """)

        store = Store(linker.engine)
        store.set_wasi(WasiConfig())
        linker.instantiate(store, module)

    def preopen_nonexistent(self):
        config = WasiConfig()
        with self.assertRaises(WasmtimeError):
            config.preopen_dir('/path/to/nowhere', '/', DirPerms.READ_ONLY, FilePerms.READ_ONLY)

    def test_custom_print(self):
        linker = Linker(Engine())
        linker.define_wasi()

        stderr = ''
        stdout = ''

        def on_stdout(data: bytes) -> None:
            nonlocal stdout
            stdout += data.decode('utf8')

        def on_stderr(data: bytes) -> None:
            nonlocal stderr
            stderr += data.decode('utf8')

        module = Module(linker.engine, """
            (module
                (import "wasi_snapshot_preview1" "fd_write"
                    (func $write (param i32 i32 i32 i32) (result i32)))

                (memory (export "memory") 1)

                (func $print
                    (i32.store (i32.const 300) (i32.const 100)) ;; iov base
                    (i32.store (i32.const 304) (i32.const 14))  ;; iov len

                    (call $write
                        (i32.const 1)    ;; fd 1 is stdout
                        (i32.const 300)  ;; iovecs ptr
                        (i32.const 1)    ;; iovecs len
                        (i32.const 400)) ;; nwritten ptr
                    if unreachable end   ;; verify no error

                    (i32.store (i32.const 300) (i32.const 200)) ;; iov base
                    (i32.store (i32.const 304) (i32.const 14))  ;; iov len

                    (call $write
                        (i32.const 2)    ;; fd 2 is stderr
                        (i32.const 300)  ;; iovecs ptr
                        (i32.const 1)    ;; iovecs len
                        (i32.const 400)) ;; nwritten ptr
                    if unreachable end   ;; verify no error
                )

                (start $print)

                (data (i32.const 100) "Hello, stdout!")
                (data (i32.const 200) "Hello, stderr!")
            )
        """)

        wasi = WasiConfig()
        wasi.stdout_custom = on_stdout
        wasi.stderr_custom = on_stderr

        store = Store(linker.engine)
        store.set_wasi(wasi)
        linker.instantiate(store, module)

        self.assertEqual('Hello, stdout!', stdout)
        self.assertEqual('Hello, stderr!', stderr)
